
find_people_response = {
    'vanId': 19722445,
    'firstName': 'Bob',
    'lastName': 'Smith',
    'middleName': None,
    'suffix': None,
    'title': None,
    'contactMode': None,
    'organizationContactCommonName': None,
    'organizationContactOfficialName': None,
    'salutation': 'Bob',
    'formalSalutation': 'Bob Smith',
    'additionalSalutation': None,
    'preferredPronoun': None,
    'envelopeName': 'Bob Smith',
    'formalEnvelopeName': 'Bob Smith',
    'additionalEnvelopeName': None,
    'contactMethodPreferenceCode': None,
    'nickname': None,
    'website': None,
    'professionalSuffix': None,
    'party': None,
    'employer': None,
    'occupation': None,
    'sex': 'F',
    'dateOfBirth': '1975-09-18T00:00:00Z',
    'selfReportedRace': None,
    'selfReportedEthnicity': None,
    'selfReportedRaces': None,
    'selfReportedEthnicities': None,
    'selfReportedGenders': None,
    'selfReportedSexualOrientations': None,
    'selfReportedLanguagePreference': None,
    'emails': None,
    'phones': None,
    'addresses': None,
    'recordedAddresses': None,
    'identifiers': None,
    'codes': None,
    'customFields': None,
    'contributionSummary': None,
    'suppressions': None,
    'caseworkCases': None,
    'caseworkIssues': None,
    'caseworkStories': None,
    'notes': None,
    'scores': None,
    'customProperties': None,
    'electionRecords': None,
    'membershipStatus': None,
    'organizationRoles': None,
    'districts': None
}

get_person_response = {
    'vanId': 19722445,
    'firstName': 'Bob',
    'lastName': 'Smith',
    'middleName': None,
    'suffix': None,
    'title': None,
    'contactMode': None,
    'organizationContactCommonName': None,
    'organizationContactOfficialName': None,
    'salutation': 'Bob',
    'formalSalutation': 'Bob Smith',
    'additionalSalutation': None,
    'preferredPronoun': None,
    'envelopeName': 'Bob Smith',
    'formalEnvelopeName': 'Bob Smith',
    'additionalEnvelopeName': None,
    'contactMethodPreferenceCode': None,
    'nickname': None,
    'website': None,
    'professionalSuffix': None,
    'party': None,
    'employer': None,
    'occupation': None,
    'sex': 'F',
    'dateOfBirth': '1975-09-18T00:00:00Z',
    'selfReportedRace': None,
    'selfReportedEthnicity': None,
    'selfReportedRaces': None,
    'selfReportedEthnicities': None,
    'selfReportedGenders': None,
    'selfReportedSexualOrientations': None,
    'selfReportedLanguagePreference': None,
    'emails': [],
    'phones': [{
        'phoneId': 15406767,
        'phoneNumber': '4142020792',
        'ext': None,
        'dateCreated': '2019-01-11T12:19:00Z',
        'phoneType': 'Cell',
        'isPreferred': True,
        'smsOptInStatus': 'Unknown',
        'phoneOptInStatus': 'Unknown',
        'isCellStatus': {
            'statusId': 1,
            'statusName': 'Verified Cell'
        }
    }],
    'addresses': [{
        'addressId': None,
        'addressLine1': '900 N Washtenaw Ave',
        'addressLine2': None,
        'addressLine3': None,
        'city': 'Chicago',
        'stateOrProvince': 'IL',
        'zipOrPostalCode': '60622-4455',
        'geoLocation': None,
        'countryCode': 'US',
        'preview': '900 N Washtenaw Ave ',
        'type': 'Voting',
        'isPreferred': True,
        'streetAddress': '900 N Washtenaw Ave',
        'displayMode': 'Standardized'
    }, {
        'addressId': None,
        'addressLine1': '900 N Washtenaw Ave',
        'addressLine2': None,
        'addressLine3': None,
        'city': 'Chicago',
        'stateOrProvince': 'IL',
        'zipOrPostalCode': '60622-4455',
        'geoLocation': None,
        'countryCode': 'US',
        'preview': '900 N Washtenaw Ave ',
        'type': 'Mailing',
        'isPreferred': None,
        'streetAddress': '900 N Washtenaw Ave',
        'displayMode': 'Standardized'
    }],
    'recordedAddresses': [],
    'identifiers': [{
        'type': 'StateFileID',
        'externalId': 'KLYW682Z'
    }],
    'codes': None,
    'customFields': [],
    'contributionSummary': None,
    'suppressions': [],
    'caseworkCases': None,
    'caseworkIssues': None,
    'caseworkStories': None,
    'notes': None,
    'scores': None,
    'customProperties': None,
    'electionRecords': None,
    'membershipStatus': None,
    'organizationRoles': None,
    'districts': None
}



import pytest
from parsons import Table


'''
Simple Table

The bare minimum Parsons table, and matching files representing that table.
'''

@pytest.fixture
def simple_table():
    # Note - If you modify this table, you must also update the related "simple" files.
    # Fortunately Parsons should make that easy to do :)
    return Table([{'first': 'Bob', 'last': 'Smith'}])

@pytest.fixture
def simple_csv_path(shared_datadir):
    return str(shared_datadir / 'test-simple-table.csv')

@pytest.fixture
def simple_compressed_csv_path(shared_datadir):
    return str(shared_datadir / 'test-simple-table.csv.gz')



from contextlib import contextmanager
import paramiko
from parsons.utilities import files


class SFTP(object):
    """
    Instantiate SFTP Class

    `Args:`
        host: str
            The host name
        username: str
            The user name
        password: str
            The password
        port: int
            Specify if different than the standard port 22
    `Returns:`
        SFTP Class
    """

    def __init__(self, host, username, password, port=22):
        self.host = host
        if not self.host:
            raise ValueError("Missing the SFTP host name")

        self.username = username
        if not self.username:
            raise ValueError("Missing the SFTP username")

        self.password = password
        self.port = port

    @contextmanager
    def _create_connection(self):
        # Internal method to create a connection and close when it's out of scope.
        # Make sure to use this in a ``with`` block, since it's a context manager.

        transport = paramiko.Transport((self.host, self.port))
        transport.connect(username=self.username, password=self.password)
        conn = paramiko.SFTPClient.from_transport(transport)
        yield conn
        conn.close()
        transport.close()

    def list_directory(self, remote_path='.'):
        """
        List the contents of a directory

        `Args:`
            remote_path: str
                The remote path of the directory
        `Returns:`
            list
        """

        with self._create_connection() as conn:
            file_list = conn.listdir(path=remote_path)

        return file_list

    def make_directory(self, remote_path):
        """
        Makes a new directory on the SFTP server

        `Args:`
            remote_path: str
                The remote path of the directory
        """

        with self._create_connection() as conn:
            conn.mkdir(remote_path)

    def remove_directory(self, remote_path):
        """
        Remove a directory from the SFTP server

        `Args:`
            remote_path: str
                The remote path of the directory
        """

        with self._create_connection() as conn:
            conn.rmdir(remote_path)

    def get_file(self, remote_path, local_path=None):
        """
        Download a file from the SFTP server

        `Args:`
            remote_path: str
                The remote path of the file to download
            local_path: str
                The local path where the file will be downloaded. If not specified, a temporary
                file will be created and returned, and that file will be removed automatically
                when the script is done running.

        `Returns:`
            str
                The path of the local file
        """

        if not local_path:
            local_path = files.create_temp_file_for_path(remote_path)

        with self._create_connection() as conn:
            conn.get(remote_path, local_path)

        return local_path

    def put_file(self, local_path, remote_path):
        """
        Put a file on the SFTP server
        `Args:`
            local_path: str
                The local path of the source file
            remote_path: str
                The remote path of the new file
        """
        with self._create_connection() as conn:
            conn.put(local_path, remote_path)

    def remove_file(self, remote_path):
        """
        Delete a file on the SFTP server

        `Args:`
            remote_path: str
                The remote path of the file
        """

        with self._create_connection() as conn:
            conn.remove(remote_path)

    def get_file_size(self, remote_path):
        """
        Get the size of a file in MB on the SFTP Server

        `Args:`
            remote_path: str
                The remote path of the file
        `Returns:`
            int
                The file size in MB.
        """

        with self._create_connection() as conn:
            size = conn.file(remote_path, 'r')._get_size()

        return size / 1024



"""NGPVAN Supporter Groups Endpoints"""
from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class SupporterGroups(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_supporter_groups(self):
        """
        Get supporter groups.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('supporterGroups'))
        logger.info(f'Found {tbl.num_rows} supporter groups.')
        return tbl

    def get_supporter_group(self, supporter_group_id):
        """
        Get a supporter group.

        `Args:`
            supporter_group_id: int
                The supporter group id.
        `Returns:`
            dict
        """

        r = self.connection.get_request(f'supporterGroups/{supporter_group_id}')
        logger.info(f'Found supporter group {supporter_group_id}.')
        return r

    def create_supporter_group(self, name, description):
        """
        Create a new supporter group

        `Args:`
            name: str
                The name of the supporter group. 100 character limit
            description: str
                Optional; A description of the supporter group. 200 character limit.
        `Returns`
            Parsons Table with the newly createed supporter group id, name
            and description
        """

        json = {'name': name, 'description': description}
        r = self.connection.post_request('supporterGroups', json=json)
        return r

    def add_person_supporter_group(self, supporter_group_id, vanid):
        """
        Add a person to a supporter group

        `Args:`
            supporter_group_id: int
                The supporter group id
            vanid: int
                The vanid of the person to apply
        `Returns:`
            ``None``
        """

        r = self.connection.put_request(f'supporterGroups/{supporter_group_id}/people/{vanid}')
        logger.info(f'Added person {vanid} to {supporter_group_id} supporter group.')
        return r

    def delete_person_supporter_group(self, supporter_group_id, vanid):
        """
        Remove a person from a supporter group

        `Args:`
            supporter_group_id: int
                The supporter group id
            vanid: int
                The vanid of the person to remove
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f'supporterGroups/{supporter_group_id}/people/{vanid}')
        logger.info(f'Deleted person {vanid} from {supporter_group_id} supporter group.')
        return r



from google.cloud.bigquery import Client
from parsons import Table
from parsons.google.utitities import setup_google_application_credentials
from parsons.utilities.files import create_temp_file
import petl
import pickle


class GoogleBigQuery:
    """
    Class for querying BigQuery table and returning the data as Parsons tables.

    This class requires application credentials in the form of a json. It can be passed
    in the following ways:

    * Set an environmental variable named ``GOOGLE_APPLICATION_CREDENTIALS`` with the
      local path to the credentials json.

      Example: ``GOOGLE_APPLICATION_CREDENTALS='path/to/creds.json'``

    * Pass in the path to the credentials using the ``app_creds`` argument.

    * Pass in a json string using the ``app_creds`` argument.

    Args:
        project_id: str
            The project which the client is acting on behalf of. If not passed
            then will use the default inferred environment.
        app_creds: str
            A credentials json string or a path to a json file. Not required
            if ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set.
        location: str
            Default geographic location for tables
    """

    def __init__(self, app_creds=None, project=None, location=None):
        setup_google_application_credentials(app_creds)

        self.project = project
        self.location = location

        # We will not create the client until we need to use it, since creating the client
        # without valid GOOGLE_APPLICATION_CREDENTIALS raises an exception.
        # This attribute will be used to hold the client once we have created it.
        self._client = None

    def query(self, sql):
        """
        Run a BigQuery query and return the results as a Parsons table.

        `Args:`
            sql: str
                A valid BigTable statement

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        # Run the query
        query_job = self.client.query(sql)

        # We will use a temp file to cache the results so that they are not all living
        # in memory. We'll use pickle to serialize the results to file in order to maintain
        # the proper data types (e.g. integer).
        temp_filename = create_temp_file()

        wrote_header = False
        with open(temp_filename, 'wb') as temp_file:
            results = query_job.result()

            # If there are no results, just return None
            if results.total_rows == 0:
                return None

            for row in results:
                # Make sure we write out the header once and only once
                if not wrote_header:
                    wrote_header = True
                    header = list(row.keys())
                    pickle.dump(header, temp_file)

                row_data = list(row.values())
                pickle.dump(row_data, temp_file)

        ptable = petl.frompickle(temp_filename)
        final_table = Table(ptable)

        return final_table

    @property
    def client(self):
        """
        Get the Google BigQuery client to use for making queries.

        `Returns:`
            `google.cloud.bigquery.client.Client`
        """
        if not self._client:
            # Create a BigQuery client to use to make the query
            self._client = Client(project=self.project, location=self.location)

        return self._client



import unittest
import requests_mock
from parsons.targetsmart.targetsmart_api import TargetSmartAPI
from parsons.etl.table import Table
from test.utils import validate_list
from test.responses.ts_responses import *

output_list = [{'vb.tsmart_zip': '60625',
                'vb.vf_g2014': 'Y',
                'vb.vf_g2016': 'Y',
                'vb.tsmart_middle_name': 'H',
                'ts.tsmart_midterm_general_turnout_score': '85.5',
                'vb.tsmart_name_suffix': '',
                'vb.voterbase_gender': 'Male',
                'vb.tsmart_city': 'CHICAGO',
                'vb.tsmart_full_address': '908 N MAIN AVE APT 2',
                'vb.voterbase_phone': '5125705356',
                'vb.tsmart_partisan_score': '99.6',
                'vb.tsmart_last_name': 'BLANKS',
                'vb.voterbase_id': 'IL-12568670',
                'vb.tsmart_first_name': 'BILLY',
                'vb.voterid': 'Q8W8R82Z',
                'vb.voterbase_age': '37',
                'vb.tsmart_state': 'IL',
                'vb.voterbase_registration_status': 'Registered'
                }]


class TestTargetSmartAPI(unittest.TestCase):

    def setUp(self):

        self.ts = TargetSmartAPI(api_key='FAKEKEY')

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_data_enhance(self, m):

        json = {'input':
                {'search_id': 'IL-12568670',
                 'search_id_type': 'voterbase'},
                'error': None,
                'output': output_list,
                'output_size': 1,
                'match_found': True,
                'gateway_id': 'b8c86f27-fb32-11e8-9cc1-45bc340a4d22',
                'function_id': 'b8c98093-fb32-11e8-8b25-e99c70f6fe74'}

        expected = ['vb.tsmart_zip', 'vb.vf_g2014', 'vb.vf_g2016', 'vb.tsmart_middle_name',
                    'ts.tsmart_midterm_general_turnout_score', 'vb.tsmart_name_suffix',
                    'vb.voterbase_gender', 'vb.tsmart_city', 'vb.tsmart_full_address',
                    'vb.voterbase_phone', 'vb.tsmart_partisan_score', 'vb.tsmart_last_name',
                    'vb.voterbase_id', 'vb.tsmart_first_name', 'vb.voterid', 'vb.voterbase_age',
                    'vb.tsmart_state', 'vb.voterbase_registration_status']

        m.get(self.ts.connection.uri + 'person/data-enhance', json=json)

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.ts.data_enhance('IL-12568678')))

        # Assert exception on missing state
        with self.assertRaises(Exception):
            self.ts.data_enhance('vb0001', search_id_type='votebuilder')

        # Assert exception on missing state
        with self.assertRaises(Exception):
            self.ts.data_enhance('vb0001', search_id_type='smartvan')

        # Assert exception on missing state
        with self.assertRaises(Exception):
            self.ts.data_enhance('vb0001', search_id_type='voter')

        # Assert works with state provided
        for i in ['votebuilder', 'voter', 'smartvan']:
            self.assertTrue(validate_list(expected, self.ts.data_enhance(
                'IL-12568678',
                search_id_type=i,
                state='IL')))

    @requests_mock.Mocker()
    def test_radius_search(self, m):

        m.get(self.ts.connection.uri + 'person/radius-search', json=radius_response)

        expected = ['similarity_score', 'distance_km', 'distance_meters', 'distance_miles',
                    'distance_feet', 'proximity_score', 'composite_score', 'uniqueness_score',
                    'confidence_indicator', 'ts.tsmart_midterm_general_turnout_score',
                    'vb.tsmart_city', 'vb.tsmart_first_name', 'vb.tsmart_full_address',
                    'vb.tsmart_last_name', 'vb.tsmart_middle_name', 'vb.tsmart_name_suffix',
                    'vb.tsmart_partisan_score', 'vb.tsmart_precinct_id', 'vb.tsmart_precinct_name',
                    'vb.tsmart_state', 'vb.tsmart_zip', 'vb.tsmart_zip4',
                    'vb.vf_earliest_registration_date', 'vb.vf_g2014', 'vb.vf_g2016',
                    'vb.vf_precinct_id', 'vb.vf_precinct_name', 'vb.vf_reg_cass_address_full',
                    'vb.vf_reg_cass_city', 'vb.vf_reg_cass_state', 'vb.vf_reg_cass_zip',
                    'vb.vf_reg_cass_zip4', 'vb.vf_registration_date', 'vb.voterbase_age',
                    'vb.voterbase_gender', 'vb.voterbase_id', 'vb.voterbase_phone',
                    'vb.voterbase_registration_status', 'vb.voterid']

        # Assert response is expected structure
        def rad_search():
            return self.ts.radius_search('BILLY', 'Burchard', radius_size=100,
                                         address='908 N Washtenaw, Chicago, IL')

        self.assertTrue(validate_list(expected, rad_search()))

    def test_district_args(self):

    	self.assertRaises(ValueError, self.ts.district, search_type='address')
    	self.assertRaises(ValueError, self.ts.district, search_type='zip', zip4=9)
    	self.assertRaises(ValueError, self.ts.district, search_type='zip', zip5=0)
    	self.assertRaises(ValueError, self.ts.district, search_type='point')
    	self.assertRaises(ValueError, self.ts.district, search_type='zip')

    @requests_mock.Mocker()
    def test_district_point(self, m):

        # Test Points
        m.get(self.ts.connection.uri + 'service/district', json=district_point)
        self.assertTrue(validate_list(district_expected,
                                      self.ts.district(search_type='point',
                                                       latitude='41.898369',
                                                       longitude='-87.694382')))

    @requests_mock.Mocker()
    def test_district_zip(self, m):
        # Test Zips
        m.get(self.ts.connection.uri + 'service/district', json=district_zip)
        self.assertTrue(validate_list(zip_expected,
                                      self.ts.district(search_type='zip',
                                                       zip5='60622',
                                                       zip4='7194')))
    @requests_mock.Mocker()
    def test_district_address(self, m):
        # Test Address
        m.get(self.ts.connection.uri + 'service/district', json=address_response)
        self.assertTrue(validate_list(district_expected,
                                      self.ts.district(search_type='address',
                                                       address='908 N Main St, Chicago, IL 60611')))

    @requests_mock.Mocker()
    def test_phone(self, m):

        # Test phone
        m.get(self.ts.connection.uri + 'person/phone-search', json=phone_response)
        self.assertTrue(validate_list(phone_expected,
                                      self.ts.phone(Table([{'phone': 4435705355}]))))



import re
import boto3
from parsons.utilities import files
import logging

logger = logging.getLogger(__name__)


class AWSConnection(object):

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None):

        # Order of operations for searching for keys:
        #   1. Look for keys passed as kwargs
        #   2. Look for env variables
        #   3. Look for aws config file
        # Boto3 handles 2 & 3, but should handle 1 on it's own. Not sure
        # why that's not working.

        if aws_access_key_id and aws_secret_access_key:

            self.session = boto3.Session(aws_access_key_id=aws_access_key_id,
                                         aws_secret_access_key=aws_secret_access_key)

        else:
            self.session = boto3.Session()


class S3(object):
    """
    Instantiate the S3 class.

    `Args:`
        aws_access_key_id: str
            The AWS access key id. Not required if the ``AWS_ACCESS_KEY_ID`` env variable
            is set.
        aws_secret_access_key: str
            The AWS secret access key. Not required if the ``AWS_SECRET_ACCESS_KEY`` env
            variable is set.
    `Returns:`
        S3 class.
    """

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None):

        self.aws = AWSConnection(aws_access_key_id=aws_access_key_id,
                                 aws_secret_access_key=aws_secret_access_key)

        self.s3 = self.aws.session.resource('s3')
        """Boto3 API Session Resource object. Use for more advanced boto3 features."""

        self.client = self.s3.meta.client
        """Boto3 API Session client object. Use for more advanced boto3 features."""

    def list_buckets(self):
        """
        List all buckets to which you have access.

        `Returns:`
            list
        """

        return [bucket.name for bucket in self.s3.buckets.all()]

    def bucket_exists(self, bucket):
        """
        Determine if a bucket exists and you have access to it.

        `Args:`
            bucket: str
                The bucket name
        `Returns:`
            boolean
                ``True`` if the bucket exists and ``False`` if not.
        """

        try:
            # If we can list the keys, the bucket definitely exists. We do this check since
            # it will account for buckets that live on other AWS accounts and that we
            # have access to.
            self.list_keys(bucket)
            return True
        except Exception:
            pass

        return bucket in self.list_buckets()

    def list_keys(self, bucket, prefix=None, suffix=None, regex=None,
                  date_modified_before=None, date_modified_after=None):
        """
        List the keys in a bucket, along with extra info about each one.

        `Args:`
            bucket: str
                The bucket name
            prefix: str
                Limits the response to keys that begin with the specified prefix.
            suffix: str
                Limits the response to keys that end with specified suffix
            regex: str
                Limits the reponse to keys that match a regex pattern
            date_modified_before: datetime.datetime
                Limits the response to keys with date modified before
            date_modified_after: datetime.datetime
                Limits the response to keys with date modified after
        `Returns:`
            dict
                Dict mapping the keys to info about each key. The info includes 'LastModified',
                'Size', and 'Owner'.
        """

        keys_dict = dict()
        logger.info(f'Fetching keys in {bucket} bucket')

        continuation_token = None

        while True:
            args = {'Bucket': bucket}
            if prefix:
                args['Prefix'] = prefix
            if continuation_token:
                args['ContinuationToken'] = continuation_token

            resp = self.client.list_objects_v2(**args)

            for key in resp.get('Contents', []):

                # Match suffix
                if suffix and not key['Key'].endswith(suffix):
                    continue

                # Regex matching
                if regex and not bool(re.search(regex, key['Key'])):
                    continue

                # Match timestamp parsing
                if date_modified_before and not key['LastModified'] < date_modified_before:
                    continue

                if date_modified_after and not key['LastModified'] > date_modified_after:
                    continue

                # Convert date to iso string
                key['LastModified'] = key['LastModified'].isoformat()

                # Add to output dict
                keys_dict[key.get('Key')] = key

            # If more than 1000 results, continue with token
            if resp.get('NextContinuationToken'):
                continuation_token = resp['NextContinuationToken']
            else:
                break

        logger.info(f'Retrieved {len(keys_dict)} keys')
        return keys_dict

    def key_exists(self, bucket, key):
        """
        Determine if a key exists in a bucket.

        `Args:`
            bucket: str
                The bucket name
            key: str
                The object key
        `Returns:`
            boolean
                ``True`` if key exists and ``False`` if not.
        """

        key_count = len(self.list_keys(bucket, prefix=key))

        if key_count > 0:
            logger.info(f'Found {key} in {bucket}.')
            return True
        else:
            logger.info(f'Did not find {key} in {bucket}.')
            return False

    def create_bucket(self, bucket):
        """
        Create an s3 bucket.

        .. warning::
            S3 has a limit on the number of buckets you can create in an AWS account, and
            that limit is fairly low (typically 100). If you are creating buckets frequently,
            you may be mis-using S3, and should consider using the same bucket for multiple tasks.
            There is no limit on the number of objects in a bucket.
            See `AWS bucket restrictions
            <https://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html>`_ for more
            info.

        .. warning::
            S3 bucket names are *globally* unique. So when creating a new bucket,
            the name can't collide with any existing bucket names. If the provided name does
            collide, you'll see errors like `IllegalLocationConstraintException` or
            `BucketAlreadyExists`.

        `Args:`
            bucket: str
                The name of the bucket to create
        `Returns:`
            ``None``
        """

        self.client.create_bucket(Bucket=bucket)

    def put_file(self, bucket, key, local_path, acl='bucket-owner-full-control', **kwargs):
        """
        Uploads an object to an S3 bucket

        `Args:`
            bucket: str
                The bucket name
            key: str
                The object key
            local_path: str
                The local path of the file to upload
            acl: str
                The S3 permissions on the file
            kwargs:
                Additional arguments for the S3 API call. See `AWS Put Object documentation
                <https://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectPUT.html>`_ for more
                info.
        """

        self.client.upload_file(local_path, bucket, key, ExtraArgs={'ACL': acl, **kwargs})

    def remove_file(self, bucket, key):
        """
        Deletes an object from an S3 bucket

        `Args:`
            bucket: str
                The bucket name
            key: str
                The object key
        `Returns:`
            ``None``
        """

        self.client.delete_object(Bucket=bucket, Key=key)

    def get_file(self, bucket, key, local_path=None):
        """
        Download an object from S3 to a local file

        `Args:`
            local_path: str
                The local path where the file will be downloaded. If not specified, a temporary
                file will be created and returned, and that file will be removed automatically
                when the script is done running.
            bucket: str
                The bucket name
            key: str
                The object key

        `Returns:`
            str
                The path of the new file
        """

        if not local_path:
            local_path = files.create_temp_file_for_path(key)

        self.s3.Object(bucket, key).download_file(local_path)

        return local_path

    def get_url(self, bucket, key, expires_in=3600):
        """
        Generates a presigned url for an s3 object.

        `Args:`
            bucket: str
                The bucket name
            key: str
                The object name
            expires_in: int
                The time, in seconds, until the url expires
        `Returns:`
            Url:
                A link to download the object
        """

        return self.client.generate_presigned_url(ClientMethod='get_object',
                                                  Params={'Bucket': bucket,
                                                          'Key': key},
                                                  ExpiresIn=expires_in)

    def transfer_bucket(self, origin_bucket, origin_key, destination_bucket,
                        destination_key=None, suffix=None, regex=None,
                        date_modified_before=None, date_modified_after=None,
                        public_read=False):
        """Transfer files between s3 buckets
        `Args:`
            origin_bucket: str
                The origin bucket
            origin_key: str
                The origin file or prefix
            destination_bucket: str
                The destination bucket
            destination_key: str
                If `None` then will retain the `origin key`. If set to prefix will move all
                to new prefix
            suffix: str
                Limits the response to keys that end with specified suffix
            regex: str
                Limits the reponse to keys that match a regex pattern
            date_modified_before: datetime.datetime
                Limits the response to keys with date modified before
            date_modified_after: datetime.datetime
                Limits the response to keys with date modified after
            public_read: bool
                If the keys should be set to `public-read`
        `Returns:`
            ``None``
        """
        # If prefix, get all files for the prefix
        if origin_key.endswith('/'):
            resp = self.list_keys(
                origin_bucket,
                prefix=origin_key,
                suffix=suffix,
                regex=regex,
                date_modified_before=date_modified_before,
                date_modified_after=date_modified_after
            )
            key_list = [value['Key'] for value in resp.values()]
        else:
            key_list = [origin_key]

        for key in key_list:
            # If destination_key is prefix, replace
            if destination_key and destination_key.endswith('/'):
                dest_key = key.replace(origin_key, destination_key)

            # If single destination, use destination key
            elif destination_key:
                dest_key = destination_key

            # Else use key from original source
            else:
                dest_key = key

            copy_source = {'Bucket': origin_bucket, 'Key': key}
            self.client.copy(copy_source, destination_bucket, dest_key)

            if public_read:
                object_acl = self.s3.ObjectAcl(destination_bucket, destination_key)
                object_acl.put(ACL='public-read')

        logger.info(f'Finished syncing {len(key_list)} keys')



import unittest
import os
import pytest
import shutil
from unittest import mock
from parsons.etl.table import Table
from parsons.utilities import date_convert
from parsons.utilities import files
from parsons.utilities import check_env
from parsons.utilities import json_format


"""
# Does not work locally due to some UTC issues, but works on CircleCI. Commenting
# out for the time being.

class TestDateConvert(unittest.TestCase):

    def test_date_convert(self):

        self.assertEqual(date_convert.iso_to_unix('2018-12-13'), 1544659200)
"""

#
# File utility tests (pytest-style)
#


def test_create_temp_file_for_path():
    temp_path = files.create_temp_file_for_path('some/file.gz')
    assert temp_path[-3:] == '.gz'

def test_close_temp_file():
    temp = files.create_temp_file()
    files.close_temp_file(temp)

    # Verify the temp file no longer exists
    with pytest.raises(FileNotFoundError):
        open(temp, 'r')


def test_is_gzip_path():
    assert files.is_gzip_path('some/file.gz')
    assert not files.is_gzip_path('some/file')
    assert not files.is_gzip_path('some/file.csv')


def test_suffix_for_compression_type():
    assert files.suffix_for_compression_type(None) == ''
    assert files.suffix_for_compression_type('') == ''
    assert files.suffix_for_compression_type('gzip') == '.gz'


def test_compression_type_for_path():
    assert files.compression_type_for_path('some/file') == None
    assert files.compression_type_for_path('some/file.csv') == None
    assert files.compression_type_for_path('some/file.csv.gz') == 'gzip'

def test_empty_file():

    # Create fake files.
    os.mkdir('tmp')
    with open('tmp/empty.csv', 'w+') as f:
        pass
    Table([['1'],['a']]).to_csv('tmp/full.csv')

    assert files.has_data('tmp/empty.csv') == False
    assert files.has_data('tmp/full.csv') == True

    # Remove fake files and dir
    shutil.rmtree('tmp')

def test_json_format():
    assert json_format.arg_format('my_arg') == 'myArg'

def test_remove_empty_keys():

    # Assert key removed when None
    test_dict = {'a': None, 'b': 2}
    assert json_format.remove_empty_keys(test_dict) == {'b': 2}

    # Assert key not removed when None
    test_dict = {'a': 1, 'b': 2}
    assert json_format.remove_empty_keys(test_dict) == {'a': 1, 'b': 2}

    # Assert that a nested empty string is removed
    test_dict = {'a': '', 'b': 2}
    assert json_format.remove_empty_keys(test_dict) ==  {'b': 2}

class TestCheckEnv(unittest.TestCase):

    def test_environment_field(self):
        """Test check field"""
        result = check_env.check('PARAM', 'param')
        self.assertEqual(result, 'param')

    @mock.patch.dict(os.environ, {'PARAM': 'env_param'})
    def test_environment_env(self):
        """Test check env"""
        result = check_env.check('PARAM', None)
        self.assertEqual(result, 'env_param')

    @mock.patch.dict(os.environ, {'PARAM': 'env_param'})
    def test_environment_field_env(self):
        """Test check field with env and field"""
        result = check_env.check('PARAM', 'param')
        self.assertEqual(result, 'param')

    def test_envrionment_error(self):
        """Test check env raises error"""
        with self.assertRaises(KeyError) as context:
            check_env.check('PARAM', None)




"""Mobile Commons Campaings Endpoints."""


class Campaigns(object):
    """Class for campaigns endpoints."""

    def __init__(self, mc_connection):
        """Initialize the Campaigns class.

        `Args:`
            mc_connection: MobileCommonsConnector
                The connector to access the Mobile Commons API.
        """
        self.connection = mc_connection

    def campaigns(self, include_opt_in_paths=0, sort='asc', status=None,
                  campaign_id=None):
        """Return a list of campaigns.

        `Args:`
            include_opt_in_paths: int
                Set to 1 to include all opt-in path details. Default is 0.
            sort: str
                Set to `asc` or `desc` to sort by campaign ID ascending or
                descending. Default is ascending.
            status: str
                Set to active or ended to filter results. Default is empty and
                returns all campaigns.
            campaign_id: str
                Provide a specific campaign ID to view single result, invalid
                campaign ID will return all campaigns.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'campaigns'

        args = {'include_opt_in_paths': include_opt_in_paths,
                'sort': sort,
                'status': status,
                'campaign_id': campaign_id}

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            return self.connection.output(
                response['response']['campaigns']['campaign'])
        else:
            return None

    def campaign(self, campaign_id, include_opt_in_paths=0):
        """Return a single campaign.

        `Args:`
            campaign_id: str
                Provide a specific campaign ID to view single result, invalid
                campaign ID will return all campaigns.
            include_opt_in_paths: int
                Set to 1 to include all opt-in path details. Default is 0.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'campaigns'

        args = {'include_opt_in_paths': include_opt_in_paths,
                'campaign_id': campaign_id}

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            return self.connection.output(
                response['response']['campaigns']['campaign'])
        else:
            return None



"""Mobile Commons Groups Endpoints."""


class Groups(object):
    """Class for groups endpoints."""

    def __init__(self, mc_connection):
        """Initialize the Groups class.

        `Args:`
            mc_connection: MobileCommonsConnector
                The connector to access the Mobile Commons API.
        """
        self.connection = mc_connection

    def groups(self):
        """Return a list of groups.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'groups'

        response = self.connection.request(url, resp_type='xml')

        if response['response']['success'] == 'true':
            if response['response']['groups']:
                return self.connection.output(
                    response['response']['groups']['group'])
            else:
                return None
        else:
            return None

    def group_members(self, group_id, limit=None, page=None, from_date=None,
                      to_date=None):
        """Return a list of members in a group.

        `Args:`
            group_id: int
                Required; The primary key of the group.
            limit: int
                Optional; Limits the number of returned profiles. Maximum of
                1000.
            page: int
                Optional; Specifies which page, of the total number of pages of
                results, to return.
            from_date: str
                Optional; Limits the returned profiles to ones updated after or
                on this date time. ISO-8601 format.
            to_date: str
                Optional; Limits the returned profiles to ones updated before
                or on this date time. ISO-8601 format.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'group_members'

        args = {'group_id': group_id,
                'limit': limit,
                'page': page,
                'from': from_date,
                'to': to_date}

        # TODO check with MC about pagination

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            if response['response']['group']:
                return self.connection.output(
                    response['response']['group']['profile'])
            else:
                return None
        else:
            return None

    def group_create(self, name):
        """Create a group.

        `Args:`
            name: str
                Required; The name for the new group.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'create_group'

        args = {'name': name}

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            if response['response']['group']:
                return self.connection.output(response['response']['group'])
            else:
                return None
        else:
            return None

    def group_add_members(self, group_id, phone_numbers):
        """Add a list of members to a group.

        `Args:`
            group_id: int
                Required; The primary key of the group.
            phone_numbers: list
                Required; A list of phone numbers to add to the group.
                If the phone numbers don't exist, the will be created as
                new profiles.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'add_group_member'

        if len(phone_numbers) == 0:
            raise ValueError("At least 1 phone number is required.")

        if phone_numbers:
            phone_number = ','.join(phone_numbers)

        args = {'group_id': group_id,
                'phone_number': phone_number}

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            if response['response']['group']:
                return self.connection.output(response['response']['group'])
            else:
                return None
        else:
            return None

    def group_remove_members(self, group_id, phone_number):
        """Remove a list of members from a group.

        `Args:`
            group_id: int
                Required; The primary key of the group.
            phone_number: list
                Required; A list of phone numbers to remove from the group.
                If the phone number is not a member of the group, it will
                still return the group.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'remove_group_member'

        if len(phone_number) == 0:
            raise ValueError("At least 1 phone number is required.")

        if phone_number:
            phone_number = ','.join(phone_number)

        args = {'group_id': group_id,
                'phone_number': phone_number}

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            if response['response']['group']:
                return self.connection.output(response['response']['group'])
            else:
                return None
        else:
            return None



from parsons import Table
from slackclient import SlackClient
import os


class Slack(object):

    def __init__(self, api_key=None):

        if api_key is None:

            try:
                self.api_key = os.environ["SLACK_API_TOKEN"]

            except KeyError:
                raise KeyError('Missing api_key. It must be passed as an '
                               'argument or stored as environmental variable')

        else:

            self.api_key = api_key

        self.client = SlackClient(self.api_key)

    def channels(self, fields=['id', 'name'], exclude_archived=False,
                 types=['public_channel']):
        """
        Return a list of all channels in a Slack team.

        `Args:`
            fields: list
                A list of the fields to return. By default, only the channel
                `id` and `name` are returned. See
                https://api.slack.com/methods/conversations.list for a full
                list of available fields. `Notes:` nested fields are unpacked.
            exclude_archived: bool
                Set to `True` to exclude archived channels from the list.
                Default is false.
            types: list
                Mix and match channel types by providing a list of any
                combination of `public_channel`, `private_channel`,
                `mpim` (aka group messages), or `im` (aka 1-1 messages).
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        tbl = self._paginate_request(
            "conversations.list", "channels", types=types,
            exclude_archived=exclude_archived)

        tbl.unpack_dict("topic", include_original=False, prepend=True,
                        prepend_value="topic")
        tbl.unpack_dict("purpose", include_original=False,
                        prepend=True, prepend_value="purpose")

        rm_cols = [x for x in tbl.columns if x not in fields]
        tbl.remove_column(*rm_cols)

        return tbl

    def users(self, fields=['id', 'name', 'deleted', 'profile_real_name_normalized',
                            'profile_email']):
        """
        Return a list of all users in a Slack team.

        `Args:`
            fields: list
                A list of the fields to return. By default, only the user
                `id` and `name` and `deleted` status are returned. See
                https://api.slack.com/methods/users.list for a full list of
                available fields. `Notes:` nested fields are unpacked.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._paginate_request("users.list", "members", include_locale=True)

        tbl.unpack_dict("profile", include_original=False, prepend=True,
                        prepend_value="profile")

        rm_cols = [x for x in tbl.columns if x not in fields]
        tbl.remove_column(*rm_cols)

        return tbl

    def message_channel(self, channel, text, as_user=False):
        """
        Send a message to a Slack channel

        `Args:`
            channel: str
                The name or id of a `public_channel`, a `private_channel`, or
                an `im` (aka 1-1 message).
            text: str
                Text of the message to send.
            as_user: str
                Pass true to post the message as the authenticated user,
                instead of as a bot. Defaults to false. See
                https://api.slack.com/methods/chat.postMessage#authorship for
                more information about Slack authorship.
        `Returns:`
            `dict`:
                A response json
        """
        resp = self.client.api_call(
            "chat.postMessage", channel=channel, text=text, as_user=as_user)

        return resp

    def upload_file(self, channels, filename, filetype=None,
                    initial_comment=None, title=None):
        """
        Upload a file to Slack channel(s).

        `Args:`
            channels: list
                List of channel names or IDs where the file will be shared.
            filename: str
                Name of the file to be uploaded.
            filetype: str
                A file type identifier. If None, type will be inferred base on
                file extension. This is used to determine what fields are
                available for that object. See https://api.slack.com/types/file
                for a list of valid types and for more information about the
                file object.
            initial_comment: str
                The text of the message to send along with the file.
            title: str
                Title of the file to be uploaded.
        `Returns:`
            `dict`:
                A response json
        """
        if filetype is None and '.' in filename:
            filetype = filename.split('.')[-1]

        with open(filename) as file_content:
            resp = self.client.api_call(
                "files.upload", channels=channels, file=file_content,
                filetype=filetype, initial_comment=initial_comment,
                title=title)

        return resp

    def _paginate_request(self, endpoint, collection, **kwargs):
        # The max object we're requesting at a time.
        # This is an nternal limit to not overload slack api
        LIMIT = 200

        items = []
        next_page = True
        cursor = None
        while next_page:
            users = self.client.api_call(
                endpoint, cursor=cursor, limit=LIMIT, **kwargs)

            items.extend(users[collection])

            if users["response_metadata"]["next_cursor"]:
                cursor = users["response_metadata"]["next_cursor"]
            else:
                next_page = False

        return Table(items)



"""Mobile Commons API."""

from parsons.mobile_commons.mobile_commons_connector\
    import MobileCommonsConnector
from parsons.mobile_commons.campaigns import Campaigns
from parsons.mobile_commons.profiles import Profiles
from parsons.mobile_commons.groups import Groups
import logging

logger = logging.getLogger(__name__)


class MobileCommons(Campaigns, Profiles, Groups):
    """Initialize the MobileCommons class.

    `Args:`
        username : str
            The username for a mobile commons account with API permissions.
            If it's not passed in, it will attempt to get it from the
            environement, MC_USER.
        password: str
            The password associated with the username. If it's not passed
            in, it will attempt to get it from the environement, MC_PASS.
        company: str
            Specify which account/company you want to use for this session.
            If you leave it off, your default account will be used.
        uri: str
            Base uri to make api calls.
    """

    def __init__(self, username=None, password=None, company=None,
                 uri='https://secure.mcommons.com/api/'):

        self.connection = MobileCommonsConnector(
            username=username,
            password=password,
            company=company,
            uri=uri)



import petl
import json
import io
import gzip
from parsons.utilities import files, zip_archive


class ToFrom(object):

    def to_dataframe(self, index=None, exclude=None, columns=None,
                     coerce_float=False):
        """
        Outputs table as a Pandas Dataframe

        `Args:`
            index: str, list
                Field of array to use as the index, alternately a specific set
                of input labels to use
            exclude: list
                Columns or fields to exclude
            columns: list
                Column names to use. If the passed data do not have names
                associated with them, this argument provides names for the
                columns. Otherwise this argument indicates the order of the
                columns in the result (any names not found in the data will
                become all-NA columns)
        `Returns:`
            dataframe
                Pandas DataFrame object
        """

        return petl.todataframe(self.table, index=index, exclude=exclude,
                                columns=columns, coerce_float=coerce_float)

    def to_html(self, local_path=None, encoding=None, errors='strict',
                index_header=False, caption=None, tr_style=None,
                td_styles=None, truncate=None):
        """
        Outputs table to html.

        .. warning::
                If a file already exists at the given location, it will be
                overwritten.

        `Args:`
            local_path: str
                The path to write the html locally. If not specified, a temporary file will be
                created and returned, and that file will be removed automatically when the script
                is done running.
            encoding: str
                The encoding type for `csv.writer()
                <https://docs.python.org/2/library/csv.html#csv.writer/>`_
            errors: str
                Raise an Error if encountered
            index_header: boolean
                Prepend index to column names; Defaults to False.
            caption: str
                A caption to include with the html table.
            tr_style: str or callable
                Style to be applied to the table row.
            td_styles: str, dict or callable
                Styles to be applied to the table cells.
            truncate: int
                Length of cell data.
        `Returns:`
            str
                The path of the new file
        """

        if not local_path:
            local_path = files.create_temp_file(suffix=".html")

        petl.tohtml(self.table,
                    source=local_path,
                    encoding=encoding,
                    errors=errors,
                    caption=caption,
                    index_header=index_header,
                    tr_style=tr_style,
                    td_styles=td_styles,
                    truncate=truncate)

        return local_path

    def to_csv(self, local_path=None, temp_file_compression=None, encoding=None, errors='strict',
               write_header=True, csv_name=None, **csvargs):
        """
        Outputs table to a CSV. Additional key word arguments are passed to ``csv.writer()``. So,
        e.g., to override the delimiter from the default CSV dialect, provide the delimiter
        keyword argument.

        .. warning::
                If a file already exists at the given location, it will be
                overwritten.

        `Args:`
            local_path: str
                The path to write the csv locally. If it ends in ".gz" or ".zip", the file will be
                compressed. If not specified, a temporary file will be created and returned,
                and that file will be removed automatically when the script is done running.
            temp_file_compression: str
                If a temp file is requested (ie. no ``local_path`` is specified), the compression
                type for that file. Currently "None", "gzip" or "zip" are supported.
                If a ``local_path`` is specified, this argument is ignored.
            encoding: str
                The CSV encoding type for `csv.writer()
                <https://docs.python.org/2/library/csv.html#csv.writer/>`_
            errors: str
                Raise an Error if encountered
            write_header: boolean
                Include header in output
            csv_name: str
                If ``zip`` compression (either specified or inferred), the name of csv file
                within the archive.
            \**csvargs: kwargs
                ``csv_writer`` optional arguments

        `Returns:`
            str
                The path of the new file
        """  # noqa: W605

        # If a zip archive.
        if files.zip_check(local_path, temp_file_compression):
            return self.to_zip_csv(archive_path=local_path,
                                   encoding=encoding,
                                   errors=errors,
                                   write_header=write_header,
                                   csv_name=csv_name,
                                   **csvargs)

        if not local_path:
            suffix = '.csv' + files.suffix_for_compression_type(temp_file_compression)
            local_path = files.create_temp_file(suffix=suffix)

        # Create normal csv/.gzip
        petl.tocsv(self.table,
                   source=local_path,
                   encoding=encoding,
                   errors=errors,
                   write_header=write_header,
                   **csvargs)

        return local_path

    def append_csv(self, local_path, encoding=None, errors='strict', **csvargs):
        """
        Appends table to an existing CSV.

        Additional additional key word arguments
        are passed to ``csv.writer()``. So, e.g., to override the delimiter
        from the default CSV dialect, provide the delimiter keyword argument.

        `Args:`
            local_path: str
                The local path of an existing CSV file. If it ends in ".gz", the file will
                be compressed.
            encoding: str
                The CSV encoding type for `csv.writer()
                <https://docs.python.org/2/library/csv.html#csv.writer/>`_
            errors: str
                Raise an Error if encountered
            \**csvargs: kwargs
                ``csv_writer`` optional arguments

        `Returns:`
            str
                The path of the file
        """  # noqa: W605

        petl.appendcsv(self.table,
                       source=local_path,
                       encoding=encoding,
                       errors=errors,
                       **csvargs)
        return local_path

    def to_zip_csv(self, archive_path=None, csv_name=None, encoding=None,
                   errors='strict', write_header=True, if_exists='replace', **csvargs):
        """
        Outputs table to a CSV in a zip archive. Additional key word arguments are passed to
        ``csv.writer()``. So, e.g., to override the delimiter from the default CSV dialect,
        provide the delimiter keyword argument. Use thismethod if you would like to write
        multiple csv files to the same archive.

        .. warning::
                If a file already exists in the archive, it will be overwritten.

        `Args:`
            archive_path: str
                The path to zip achive. If not specified, a temporary file will be created and
                returned, and that file will be removed automatically when the script is done
                running.
            csv_name: str
                The name of the csv file to be stored in the archive. If ``None`` will use
                the archive name.
            encoding: str
                The CSV encoding type for `csv.writer()
                <https://docs.python.org/2/library/csv.html#csv.writer/>`_
            errors: str
                Raise an Error if encountered
            write_header: boolean
                Include header in output
            if_exists: str
                If archive already exists, one of 'replace' or 'append'
            \**csvargs: kwargs
                ``csv_writer`` optional arguments

        `Returns:`
            str
                The path of the archive
        """  # noqa: W605

        if not archive_path:
            archive_path = files.create_temp_file(suffix='.zip')

        cf = self.to_csv(encoding=encoding, errors=errors, write_header=write_header, **csvargs)

        if not csv_name:
            csv_name = files.extract_file_name(archive_path, include_suffix=False) + '.csv'

        return zip_archive.create_archive(archive_path, cf, file_name=csv_name,
                                          if_exists=if_exists)

    def to_json(self, local_path=None, temp_file_compression=None, line_delimited=False):
        """
        Outputs table to a JSON file

        .. warning::
                If a file already exists at the given location, it will be
                overwritten.

        `Args:`
            local_path: str
                The path to write the JSON locally. If it ends in ".gz", it will be
                compressed first. If not specified, a temporary file will be created and returned,
                and that file will be removed automatically when the script is done running.
            temp_file_compression: str
                If a temp file is requested (ie. no ``local_path`` is specified), the compression
                type for that file. Currently "None" and "gzip" are supported.
                If a ``local_path`` is specified, this argument is ignored.
            line_delimited: bool
                Whether the file will be line-delimited JSON (with a row on each line), or a proper
                JSON file.

        `Returns:`
            str
                The path of the new file
        """

        if not local_path:
            suffix = '.json' + files.suffix_for_compression_type(temp_file_compression)
            local_path = files.create_temp_file(suffix=suffix)

        # Note we don't use the much simpler petl.tojson(), since that method reads the whole
        # table into memory before writing to file.

        if files.is_gzip_path(local_path):
            open_fn = gzip.open
            mode = 'w+t'
        else:
            open_fn = open
            mode = 'w'

        with open_fn(local_path, mode) as file:
            if not line_delimited:
                file.write('[')

            i = 0
            for row in self:
                if i:
                    if not line_delimited:
                        file.write(',')
                    file.write('\n')
                i += 1
                json.dump(row, file)

            if not line_delimited:
                file.write(']')

        return local_path

    def to_dicts(self):
        """
        Output table as a list of dicts.

        `Returns:`
            list
        """

        return list(petl.dicts(self.table))

    def to_sftp_csv(self, remote_path, host, username, password, port=22, encoding=None,
                    compression=None, errors='strict', write_header=True, **csvargs):
        """
        Writes the table to a CSV file on a remote SFTP server

        `Args:`
            remote_path: str
                The remote path of the file. If it ends in '.gz', the file will be compressed.
            host: str
                The remote host
            username: str
                The username to access the SFTP server
            password: str
                The password to access the SFTP server
            port: int
                The port number of the SFTP server
            encoding: str
                The CSV encoding type for `csv.writer()
                <https://docs.python.org/2/library/csv.html#csv.writer/>`_
            errors: str
                Raise an Error if encountered
            write_header: boolean
                Include header in output
            \**csvargs: kwargs
                ``csv_writer`` optional arguments
        """  # noqa: W605

        from parsons import SFTP
        sftp = SFTP(host, username, password, port)

        compression = files.compression_type_for_path(remote_path)

        local_path = self.to_csv(
            temp_file_compression=compression, encoding=encoding, errors=errors,
            write_header=write_header, **csvargs)
        sftp.put_file(local_path, remote_path)

    def to_s3_csv(self, bucket, key, aws_access_key_id=None,
                  aws_secret_access_key=None, compression=None, encoding=None,
                  errors='strict', write_header=True, public_url=False,
                  public_url_expires=3600, **csvargs):
        """
        Writes the table to an s3 object as a CSV

        `Args:`
            bucket: str
                The s3 bucket to upload to
            key: str
                The s3 key to name the file. If it ends in '.gz' or '.zip', the file will be
                compressed.
            aws_access_key_id: str
                Required if not included as environmental variable
            aws_secret_access_key: str
                Required if not included as environmental variable
            compression: str
                The compression type for the s3 object. Currently "None", "zip" and "gzip" are
                supported. If specified, will override the key suffix.
            encoding: str
                The CSV encoding type for `csv.writer()
                <https://docs.python.org/2/library/csv.html#csv.writer/>`_
            errors: str
                Raise an Error if encountered
            write_header: boolean
                Include header in output
            public_url: boolean
                Create a public link to the file
            public_url_expire: 3600
                The time, in seconds, until the url expires if ``public_url`` set to ``True``.
            \**csvargs: kwargs
                ``csv_writer`` optional arguments
        `Returns:`
            Public url if specified. If not ``None``.
        """  # noqa: W605

        compression = compression or files.compression_type_for_path(key)

        csv_name = files.extract_file_name(key, include_suffix=False) + '.csv'

        # Save the CSV as a temp file
        local_path = self.to_csv(temp_file_compression=compression,
                                 encoding=encoding,
                                 errors=errors,
                                 write_header=write_header,
                                 csv_name=csv_name,
                                 **csvargs)

        # Put the file on S3
        from parsons import S3
        self.s3 = S3(aws_access_key_id=aws_access_key_id,
                     aws_secret_access_key=aws_secret_access_key)
        self.s3.put_file(bucket, key, local_path)

        if public_url:
            return self.s3.get_url(bucket, key, expires_in=public_url_expires)
        else:
            return None

    def to_redshift(self, table_name, username=None, password=None, host=None,
                    db=None, port=None, **copy_args):
        """
        Write a table to a Redshift database. Note, this requires you to pass
        AWS S3 credentials or store them as environmental variables.

        Args:
            username: str
                Required if env variable ``REDSHIFT_USERNAME`` not populated
            password: str
                Required if env variable ``REDSHIFT_PASSWORD`` not populated
            host: str
                Required if env variable ``REDSHIFT_HOST`` not populated
            db: str
                Required if env variable ``REDSHIFT_DB`` not populated
            port: int
                Required if env variable ``REDSHIFT_PORT`` not populated. Port 5439 is typical.
            \**copy_args: kwargs
                See :func:`~parsons.databases.Redshift.copy`` for options.

        Returns:
            ``None``
        """  # noqa: W605

        from parsons import Redshift
        rs = Redshift(
            username=username, password=password, host=host, db=db, port=port)
        rs.copy(self, table_name, **copy_args)

    def to_petl(self):

        return self.table

    def to_civis(self, table, api_key=None, db=None, max_errors=None,
                 existing_table_rows='fail', diststyle=None, distkey=None,
                 sortkey1=None, sortkey2=None, wait=True, **civisargs):
        """
        Write the table to a Civis Redshift cluster. Additional key word
        arguments can passed to `civis.io.dataframe_to_civis()
        <https://civis-python.readthedocs.io/en/v1.9.0/generated/civis.io.dataframe_to_civis.html#civis.io.dataframe_to_civis>`_ # noqa: E501

        `Args`
            table: str
                The schema and table you want to upload to. E.g.,
                'scratch.table'. Schemas or tablenames with periods must be
                double quoted, e.g. 'scratch."my.table"'.
            api_key: str
                Your Civis API key. If not given, the CIVIS_API_KEY environment
                variable will be used.
            db: str or int
                The Civis Database. Can be database name or ID
            max_errors: int
                The maximum number of rows with errors to remove from
                the import before failing.
            diststyle: str
                The distribution style for the table. One of `'even'`, `'all'`
                or `'key'`.
            existing_table_rows: str
                The behaviour if a table with the requested name already
                exists. One of `'fail'`, `'truncate'`, `'append'` or `'drop'`.
                Defaults to `'fail'`.
            distkey: str
                The column to use as the distkey for the table.
            sortkey1: str
                The column to use as the sortkey for the table.
            sortkey2: str
                The second column in a compound sortkey for the table.
            wait: boolean
                Wait for write job to complete before exiting method.
        """

        from parsons.civis.civisclient import CivisClient
        civis = CivisClient(db=db, api_key=api_key)
        return civis.table_import(
            self, table, max_errors=None,
            existing_table_rows=existing_table_rows, diststyle=None,
            distkey=None, sortkey1=None, sortkey2=None, wait=True, **civisargs)

    @classmethod
    def from_csv(cls, local_path, **csvargs):
        """
        Create a ``parsons table`` object from a CSV file

        `Args:`
            local_path: obj
                A csv formatted local path, url or ftp. If this is a
                file path that ends in ".gz", the file will be decompressed first.
            \**csvargs: kwargs
                ``csv_reader`` optional arguments
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """  # noqa: W605

        if not files.has_data(local_path):
            raise ValueError('CSV file is empty')

        return cls(petl.fromcsv(local_path, **csvargs))

    @classmethod
    def from_csv_string(cls, str, **csvargs):
        """
        Create a ``parsons table`` object from a string representing a CSV.

        `Args:`
            str: str
                The string object to convert to a table
            **csvargs: kwargs
                ``csv_reader`` optional arguments
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        bytesio = io.BytesIO(str.encode('utf-8'))
        memory_source = petl.io.sources.MemorySource(bytesio.read())
        return cls(petl.fromcsv(memory_source, **csvargs))

    @classmethod
    def from_columns(cls, cols, header=None):
        """
        Create a ``parsons table`` from a list of lists organized as columns

        `Args:`
            cols: list
                A list of lists organized as columns
            header: list
                List of column names. If not specified, will use dummy column names
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        return cls(petl.fromcolumns(cols, header=header))

    @classmethod
    def from_json(cls, local_path, header=None, line_delimited=False):
        """
        Create a ``parsons table`` from a json file

        `Args:`
            local_path: list
                A JSON formatted local path, url or ftp. If this is a
                file path that ends in ".gz", the file will be decompressed first.
            header: list
                List of columns to use for the destination table. If omitted, columns will
                be inferred from the initial data in the file.
            line_delimited: bool
                Whether the file is line-delimited JSON (with a row on each line), or a proper
                JSON file.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if line_delimited:
            if files.is_gzip_path(local_path):
                open_fn = gzip.open
            else:
                open_fn = open

            with open_fn(local_path, 'r') as file:
                rows = [json.loads(line) for line in file]
            return cls(rows)

        else:
            return cls(petl.fromjson(local_path, header=header))

    @classmethod
    def from_redshift(cls, query, username=None, password=None, host=None,
                      db=None, port=None):
        """
        Create a ``parsons table`` from a Redshift query.

        To pull an entire Redshift table, use a query like ``SELECT * FROM tablename``.

        `Args:`
            query: str
                A valid SQL statement
            username: str
                Required if env variable ``REDSHIFT_USERNAME`` not populated
            password: str
                Required if env variable ``REDSHIFT_PASSWORD`` not populated
            host: str
                Required if env variable ``REDSHIFT_HOST`` not populated
            db: str
                Required if env variable ``REDSHIFT_DB`` not populated
            port: int
                Required if env variable ``REDSHIFT_PORT`` not populated. Port 5439 is typical.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        from parsons import Redshift
        rs = Redshift(username=username, password=password, host=host, db=db, port=port)
        return rs.query(query)

    @classmethod
    def from_s3_csv(cls, bucket, key, aws_access_key_id=None, aws_secret_access_key=None,
                    **csvargs):
        """
        Create a ``parsons table`` from a key in an S3 bucket.

        `Args:`
            bucket: str
                The S3 bucket.
            key: str
                The S3 key
            aws_access_key_id: str
                Required if not included as environmental variable.
            aws_secret_access_key: str
                Required if not included as environmental variable.
            \**csvargs: kwargs
                ``csv_reader`` optional arguments
        `Returns:`
            `parsons.Table` object
        """  # noqa: W605

        from parsons import S3
        s3 = S3(aws_access_key_id, aws_secret_access_key)
        file_obj = s3.get_file(bucket, key)

        if files.compression_type_for_path(key) == 'zip':
            file_obj = files.zip_archive.unzip_archive(file_obj)

        return cls(petl.fromcsv(file_obj, **csvargs))

    @classmethod
    def from_dataframe(cls, dataframe, include_index=False):
        """
        Create a ``parsons table`` from a Pandas dataframe.

        `Args:`
            dataframe: dataframe
                A valid Pandas dataframe objectt
            include_index: boolean
                Include index column
        """

        return cls(petl.fromdataframe(dataframe, include_index=include_index))



import unittest
import os
import requests_mock
from parsons.ngpvan.van import VAN
from parsons.etl.table import Table
from test.utils import validate_list, assert_matching_tables
from requests.exceptions import HTTPError


class TestNGPVAN(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_canvass_responses_contact_types(self, m):

        json = {"name": "Auto Dial",
                "contactTypeId": 19,
                "channelTypeName": "Phone"}

        m.get(self.van.connection.uri + 'canvassResponses/contactTypes', json=json)

        assert_matching_tables(Table(json), self.van.get_canvass_responses_contact_types())

    @requests_mock.Mocker()
    def test_get_canvass_responses_input_types(self, m):

        json = {"inputTypeId": 11, "name": "API"}
        m.get(self.van.connection.uri + 'canvassResponses/inputTypes', json=json)
        assert_matching_tables(Table(json), self.van.get_canvass_responses_input_types())

    @requests_mock.Mocker()
    def test_get_canvass_responses_result_codes(self, m):

        json = {
            "shortName": "BZ",
            "resultCodeId": 18,
            "name": "Busy",
            "mediumName": "Busy"
        }

        m.get(self.van.connection.uri + 'canvassResponses/resultCodes', json=json)
        assert_matching_tables(Table(json), self.van.get_canvass_responses_result_codes())

    @requests_mock.Mocker()
    def test_get_saved_lists(self, m):

        json = {'count': 1, 'items': [
            {"savedListId": 517612,
             "listCount": 974656,
             "name": "LikelyParents(16andunder)_DWID_S... - MN",
             "doorCount": 520709,
             "description": "null"
             }
        ], 'nextPageLink': None}

        m.get(self.van.connection.uri + 'savedLists', json=json)

        expected = ['savedListId', 'listCount', 'name', 'doorCount','description']

        self.assertTrue(validate_list(expected, self.van.get_saved_lists()))

    @requests_mock.Mocker()
    def test_get_saved_list(self, m):

        saved_list_id = 517612

        json = {"savedListId": 517612,
                "listCount": 974656,
                "name": "LikelyParents(16andunder)_DWID_S... - MN",
                "doorCount": 520709,
                "description": "null"
                }

        m.get(self.van.connection.uri + f'savedLists/{saved_list_id}', json=json)

        expected = ['savedListId', 'listCount', 'name', 'doorCount', 'description']

        self.assertEqual(self.van.get_saved_list(saved_list_id), json)

    @requests_mock.Mocker()
    def test_get_folders(self, m):

        json = {u'count': 2,
                u'items': [
                    {
                        u'folderId': 5046,
                        u'name': u'#2018_MN_active_universe'
                    },
                    {u'folderId': 2168,
                     u'name': u'API Generated Lists'
                     }
                ], u'nextPageLink': None}

        m.get(self.van.connection.uri + 'folders', json=json)

        expected = ['folderId', 'name']

        self.assertTrue(validate_list(expected, self.van.get_folders()))

    @requests_mock.Mocker()
    def test_get_folder(self, m):

        folder_id = 5046

        json = {"folderId": 5046, "name": "#2018_MN_active_universe"}

        m.get(self.van.connection.uri + f'folders/{folder_id}', json=json)

        self.assertEqual(json, self.van.get_folder(folder_id))

    @requests_mock.Mocker()
    def test_export_job_types(self, m):

        json = {u'count': 1, u'items':
                [{u'exportJobTypeId': 4, u'name': u'SavedListExport'}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'exportJobTypes', json=json)

        expected = ['exportJobTypeId', 'name']

        self.assertTrue(validate_list(expected, self.van.get_export_job_types()))

    @requests_mock.Mocker()
    def test_export_job_create(self, m):

        saved_list_id = 517612

        json = {"status": "Completed",
                "errorCode": "null",
                "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "activistCodes": "null",
                "canvassFileRequestId": 448,
                "dateExpired": "2018-09-08T16:04:00Z",
                "surveyQuestions": "null",
                "webhookUrl": "https://www.nothing.com/",
                "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",  # noqa: E501
                "savedListId": 517612,
                "districtFields": "null",
                "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "customFields": "null",
                "type": 4,
                "exportJobId": 448}

        m.post(self.van.connection.uri + 'exportJobs', json=json, status_code=201)

        expected = [
            'status',
            'errorCode',
            'exportJobGuid',
            'activistCodes',
            'canvassFileRequestId',
            'dateExpired',
            'surveyQuestions',
            'webhookUrl',
            'downloadUrl',
            'savedListId',
            'districtFields',
            'canvassFileRequestGuid',
            'customFields',
            'type',
            'exportJobId']

        self.assertEqual(json,self.van.export_job_create(saved_list_id))

    @requests_mock.Mocker()
    def test_get_export_job(self, m):

        export_job_id = 448

        json = {"status": "Completed",
                "errorCode": "null",
                "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "activistCodes": "null",
                "canvassFileRequestId": 448,
                "dateExpired": "2018-09-08T16:04:00Z",
                "surveyQuestions": "null",
                "webhookUrl": "https://www.nothing.com/",
                "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",  # noqa: E501
                "savedListId": 517612,
                "districtFields": "null",
                "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "customFields": "null",
                "type": 4,
                "exportJobId": 448}

        expected = [
            'status',
            'errorCode',
            'exportJobGuid',
            'activistCodes',
            'canvassFileRequestId',
            'dateExpired',
            'surveyQuestions',
            'webhookUrl',
            'downloadUrl',
            'savedListId',
            'districtFields',
            'canvassFileRequestGuid',
            'customFields',
            'type',
            'exportJobId']

        m.get(self.van.connection.uri + f'exportJobs/{export_job_id}', json=json)

        self.assertEqual(json, self.van.get_export_job(export_job_id))

    @requests_mock.Mocker()
    def test_get_survey_questions(self, m):

        json = {u'count': 67, u'items': [{
            "status": "Active",
            "responses": [
                {"shortName": "1",
                 "surveyResponseId": 1288926,
                 "name": "1-Strong Walz",
                         "mediumName": "1"},
                {"shortName": "2",
                 "surveyResponseId": 1288928,
                 "name": "2-Lean Walz",
                         "mediumName": "2"}],
            "scriptQuestion": "Who do you support for Governor?",
            "name": "MN Governor Gen",
                    "surveyQuestionId": 311838,
                    "mediumName": "MNGovG",
                    "shortName": "MGG",
                    "type": "Candidate",
                    "cycle": 2018
        }],
            u'nextPageLink': None}

        m.get(self.van.connection.uri + 'surveyQuestions', json=json)

        expected = ['status', 'responses', 'scriptQuestion', 'name',
                    'surveyQuestionId', 'mediumName', 'shortName',
                    'type', 'cycle']

        self.assertTrue(validate_list(expected, self.van.get_survey_questions()))

    @requests_mock.Mocker()
    def test_get_supporter_groups(self, m):

        json = {"items": [
            {
                "id": 12,
                "name": "tmc",
                "description": "A fun group."
            },
            {
                "id": 13,
                "name": "tmc",
                "description": "A fun group."
            },
        ],
            "nextPageLink": None,
            "count": 3
        }

        m.get(self.van.connection.uri + 'supporterGroups', json=json)

        expected = ['id', 'name', 'description']

        self.van.get_supporter_groups()

    @requests_mock.Mocker()
    def test_get_supporter_group(self, m):

        json = {"id": 12, "name": "tmc", "description": "A fun group."}
        m.get(self.van.connection.uri + 'supporterGroups/12', json=json)

        # Test that columns are expected columns
        self.assertEqual(self.van.get_supporter_group(12), json)

    @requests_mock.Mocker()
    def test_add_person_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_vanid = 12345
        good_uri = f'supporterGroups/{good_vanid}/people/{good_supporter_group_id}'
        m.put(self.van.connection.uri + good_uri, status_code=204)
        self.van.add_person_supporter_group(good_vanid, good_supporter_group_id)

        # Test bad input
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_uri = f'supporterGroups/{bad_vanid}/people/{bad_supporter_group_id}'
        m.put(self.van.connection.uri + bad_uri, status_code=404)
        self.assertRaises(HTTPError, self.van.add_person_supporter_group, bad_vanid, bad_supporter_group_id)

    @requests_mock.Mocker()
    def test_delete_person_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_vanid = 12345
        good_ep = f'supporterGroups/{good_vanid}/people/{good_supporter_group_id}'
        m.delete(self.van.connection.uri + good_ep, status_code=204)
        self.van.delete_person_supporter_group(good_vanid, good_supporter_group_id)

        # Test bad input raises
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_ep = f'supporterGroups/{bad_vanid}/people/{bad_supporter_group_id}'
        m.delete(self.van.connection.uri + bad_ep, status_code=404)
        self.assertRaises(HTTPError, self.van.delete_person_supporter_group, bad_vanid, bad_supporter_group_id)

if __name__ == '__main__':

    unittest.main()



import zipfile


def create_archive(archive_path, file_path, file_name=None, if_exists='replace'):
    """
    Create and fill an archive.

    `Args:`
        archive_path: str
            The file name of zip archive
        file_path: str
            The path of the file
        file_name: str
            The name of the file in the archive
        if_exists: str
            If archive already exists, one of 'replace' or 'append'
    `Returns:`
        Zip archive path
    """

    if if_exists == 'append':
        write_type = 'a'
    else:
        write_type = 'w'

    if not file_name:
        file_name = file_path.split('/')[-1]

    with zipfile.ZipFile(archive_path, write_type) as z:
        z.write(file_path, arcname=file_name, compress_type=zipfile.ZIP_STORED)

    return archive_path


def unzip_archive(archive_path):
    """
    Unzip an archive.

    `Args:`
        archive_path: str
    `Returns:`
        ``None``
    """

    with zipfile.ZipFile(archive_path, 'r') as z:
        z.extractall()



import os
import collections
import copy
import logging
from joblib import Parallel, delayed
from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.customaudience import CustomAudience
from parsons.etl.table import Table

logger = logging.getLogger(__name__)

FBKeySchema = CustomAudience.Schema.MultiKeySchema

# Max number of custom audience users we're allowed to send in one API call
MAX_FB_AUDIENCE_API_USERS = 10000


class FacebookAds(object):
    """
    A connector for Facebook Ads, eg. for creating custom audiences.

    This connector is currently a thin wrapper around the FB Business SDK, so some of that
    SDK is exposed. Eg. You may see exceptions like FacebookRequestError.

    For details of the Facebook Business SDK for Python:
    https://github.com/facebook/facebook-python-business-sdk

    `Args:`
        app_id: str
            A Facebook app ID. Required if env var FB_APP_ID is not populated.
        app_secret: str
            A Facebook app secret. Required if env var FB_APP_SECRET is not populated.
        access_token: str
            A Facebook access token. Required if env var FB_ACCESS_TOKEN is not populated.
        ad_account_id: str
            A Facebook ad account ID. Required if env var FB_AD_ACCOUNT_ID isnot populated.
    """

    # The data columns that are valid for creating a custom audience.
    # Feel free to add more variants to capture common column names, as long as they're fairly
    # unambiguous.
    # IMPORTANT - Keep these maps in sync with the comments in the ``add_users_to_custom_audience``
    # method!
    # TODO add support for parsing full names from one column
    KeyMatchMap = {
        FBKeySchema.email: ['email', 'email address', 'voterbase_email'],
        FBKeySchema.fn: ['fn', 'first', 'first name', 'vb_tsmart_first_name'],
        FBKeySchema.ln: ['ln', 'last', 'last name', 'vb_tsmart_last_name'],
        FBKeySchema.phone: [
            'phone',
            'phone number',
            'cell',
            'landline',
            'vb_voterbase_phone',
            'vb_voterbase_phone_wireless'
            ],
        FBKeySchema.ct: ['ct', 'city', 'vb_vf_reg_city', 'vb_tsmart_city'],
        FBKeySchema.st: [
            'st',
            'state',
            'state code',
            'vb_vf_source_state',
            'vb_tsmart_state',
            'vb_vf_reg_state',
            'vb_vf_reg_cass_state'
            ],
        FBKeySchema.zip: ['zip', 'zip code', 'vb_vf_reg_zip', 'vb_tsmart_zip'],
        FBKeySchema.country: ['country', 'country code'],
        # Yes, it's not kosher to confuse gender and sex. However, gender is all that FB
        # supports in their audience targeting.
        FBKeySchema.gen: ['gen', 'gender', 'sex', 'vb_voterbase_gender'],
        FBKeySchema.doby: ['doby', 'dob year', 'birth year'],
        FBKeySchema.dobm: ['dobm', 'dob month', 'birth month'],
        FBKeySchema.dobd: ['dobd', 'dob day', 'birth day'],
    }

    PreprocessKeyMatchMap = {
        # Data in this column will be parsed into the FBKeySchema.dobX keys.
        "DOB YYYYMMDD": ['dob', 'vb_voterbase_dob', 'vb_tsmart_dob']
    }

    def __init__(self, app_id=None, app_secret=None, access_token=None, ad_account_id=None):

        try:
            self.app_id = app_id or os.environ['FB_APP_ID']
            self.app_secret = app_secret or os.environ['FB_APP_SECRET']
            self.access_token = access_token or os.environ['FB_ACCESS_TOKEN']
            self.ad_account_id = ad_account_id or os.environ['FB_AD_ACCOUNT_ID']
        except KeyError as error:
            logger.error("FB Marketing API credentials missing. Must be specified as env vars "
                         "or kwargs")
            raise error

        FacebookAdsApi.init(self.app_id, self.app_secret, self.access_token)
        self.ad_account = AdAccount("act_%s" % self.ad_account_id)

    @staticmethod
    def _get_match_key_for_column(column):
        # Finds a FB match key for a given table column.

        normalized_col = Table.get_normalized_column_name(column)

        for k, match_list in FacebookAds.KeyMatchMap.items():
            for match in match_list:
                if normalized_col == Table.get_normalized_column_name(match):
                    return k
        return None

    @staticmethod
    def _get_preprocess_key_for_column(column):
        normalized_col = Table.get_normalized_column_name(column)

        for k, match_list in FacebookAds.PreprocessKeyMatchMap.items():
            for match in match_list:
                if normalized_col == Table.get_normalized_column_name(match):
                    return k
        return None

    @staticmethod
    def _preprocess_dob_column(table, column):
        # Parse the DOB column into 3 new columns, and remove the original column
        # TODO Throw an error if the values are not 6 characters long?

        table.add_column(
            FBKeySchema.doby,
            lambda row: row[column][:4] if row[column] else None
        )
        table.add_column(
            FBKeySchema.dobm,
            lambda row: row[column][4:6] if row[column] else None
        )
        table.add_column(
            FBKeySchema.dobd,
            lambda row: row[column][6:8] if row[column] else None
        )
        table.remove_column(column)

    @staticmethod
    def _preprocess_users_table(table):
        # Handle columns that require special parsing
        for column in table.columns:
            preprocess_key = FacebookAds._get_preprocess_key_for_column(column)
            if preprocess_key == "DOB YYYYMMDD":
                FacebookAds._preprocess_dob_column(table, column)
        return table

    @staticmethod
    def _get_first_non_empty_value_from_dict(dict, cols):
        for c in cols:
            if dict[c]:
                return dict[c]
        return None

    @staticmethod
    def get_match_table_for_users_table(users_table):
        """
        Prepared an input table for matching into a FB custom audience, by identifying which
        columns are supported for matching, renaming those columns to what FB expects, and
        cutting away the other columns.

        See ``FacebookAds.create_custom_audience`` for more details.

        `Args`:
            users_table: Table
                The source table for matching

        `Returns:`
            Table
                The prepared table
        """

        # Copy the table to avoid messing up the source table
        t = copy.deepcopy(users_table)

        FacebookAds._preprocess_users_table(t)

        # Map the FB keys to whatever source columns match.
        matched_cols = []
        fb_keys_to_orig_cols = collections.defaultdict(set)
        for c in t.columns:
            match_key = FacebookAds._get_match_key_for_column(c)
            if match_key:
                matched_cols.append(c)
                fb_keys_to_orig_cols[match_key].add(c)

        # Cut the table to just the columns that we can use for matching in FB
        t = t.cut(matched_cols)

        # For each of the FB match keys, create a new column from the source column.
        # If there are more than one source cols for a given FB match key, we'll pick
        # the first non-empty value for each row.

        for fb_key, orig_cols in fb_keys_to_orig_cols.items():
            value_fn = (
                lambda bound_cols:
                    lambda row:
                        FacebookAds._get_first_non_empty_value_from_dict(row, bound_cols)
            )(orig_cols)

            # A little trickery here to handle the case where one of the "orig_cols" is already
            # named like the "fb_key".
            t.add_column(fb_key+"_fb_temp_col", value_fn)
            t.remove_column(*orig_cols)
            t.rename_column(fb_key+"_fb_temp_col", fb_key)

        # Convert None values to empty strings. Otherwise the FB SDK chokes.
        petl_table = t.to_petl()
        t = Table(petl_table.replaceall(None, ''))

        return t

    @staticmethod
    def _get_match_schema_and_data(table):
        # Grab the raw data as a list of tuples
        data_list = [row for row in table.data]
        return (table.columns, data_list)

    @staticmethod
    def _is_valid_data_source(data_source):
        valid_sources = [
            CustomAudience.CustomerFileSource.user_provided_only,
            CustomAudience.CustomerFileSource.partner_provided_only,
            CustomAudience.CustomerFileSource.both_user_and_partner_provided,
        ]
        return data_source in valid_sources

    def create_custom_audience(self, name, data_source, description=None):
        """
        Creates a FB custom audience.

        `Args:`
            name: str
                The name of the custom audience
            data_source: str
                One of ``USER_PROVIDED_ONLY``, ``PARTNER_PROVIDED_ONLY``, or
                ``BOTH_USER_AND_PARTNER_PROVIDED``.
                This tells FB whether the data for a custom audience was provided by actual users,
                or acquired via partners. FB requires you to specify.
            description: str
                Optional. The description of the custom audience

        `Returns:`
            ID of the created audience
        """

        if not self._is_valid_data_source(data_source):
            raise KeyError("Invalid data_source provided")

        params = {
            'name': name,
            'subtype': 'CUSTOM',
            'description': description,
            'customer_file_source': data_source,
        }

        res = self.ad_account.create_custom_audience(params=params)
        return res['id']

    def delete_custom_audience(self, audience_id):
        """
        Deletes a FB custom audience.

        `Args:`
            audience_id: str
                The ID of the custom audience to delete.
        """

        CustomAudience(audience_id).api_delete()

    @staticmethod
    def _add_batch_to_custom_audience(app_id, app_secret, access_token, audience_id, schema,
                                      batch, added_so_far, total_rows):
        # Since this method runs in parallel, we need to re-initialize the Facebook API each time
        # to avoid SSL-related errors. Basically, the FacebookAdsApi python framework isn't
        # built to run in parallel.
        FacebookAdsApi.init(app_id, app_secret, access_token)

        # Note that the FB SDK handles basic normalization and hashing of the data
        CustomAudience(audience_id).add_users(schema, batch, is_raw=True)
        logger.info(f"Added {added_so_far+len(batch)}/{total_rows} users to custom audience...")

    def add_users_to_custom_audience(self, audience_id, users_table):
        """
        Adds user data to a custom audience.

        Each user row in the provided table should have at least one of the supported columns
        defined. Otherwise the row will be ignored. Beyond that, the rows may have any other
        non-supported columns filled out, and those will all be ignored.

        .. list-table::
            :widths: 20 80
            :header-rows: 1

            * - Column Type
              - Valid Column Names
            * - Email Address
              - ``email``, ``email address``, ``voterbase_email``
            * - First Name
              - ``fn``, ``first``, ``first name``, ``vb_tsmart_first_name``
            * - Last Name
              - ``ln``, ``last``, ``last name``, ``vb_tsmart_last_name``
            * - Phone Number
              - ``phone``, ``phone number``, ``cell``, ``landline``, ``vb_voterbase_phone``, ``vb_voterbase_phone_wireless``
            * - City
              - ``ct``, ``city``, ``vb_vf_reg_city``, ``vb_tsmart_city``
            * - State
              - ``st``, ``state``, ``state code``, ``vb_vf_source_state``, ``vb_tsmart_state``, ``vb_vf_reg_state``, ``vb_vf_reg_cass_state``
            * - Zip Code
              - ``zip``, ``zip code``, ``vb_vf_reg_zip``, ``vb_tsmart_zip``
            * - County
              - ``country``, ``country code``
            * - Gender
              - ``gen``, ``gender``, ``sex``, ``vb_vf_reg_zip``
            * - Birth Year
              - ``doby``, ``dob year``, ``birth year``
            * - Birth Month
              - ``dobm``, ``dob month``, ``birth month``
            * - Birth Day
              - ``dobd``, ``dob day``, ``birth day``
            * - Date of Birth
              - ``dob``, ``vb_voterbase_dob``, ``vb_tsmart_dob`` (Format: YYYYMMDD)

        The column names will be normalized before comparing to this list - eg. removing
        whitespace and punctuation - so you don't need to match exactly.

        If more than one of your columns map to a single FB key, then for each row we'll use any
        non-null value for those columns.
        Eg. If you have both ``vb_voterbase_phone`` and ``vb_voterbase_phone_wireless`` (which
        both map to the FB "phone" key), then for each person in your table, we'll try to pick one
        valid phone number.

        For details of the expected data formats for each column type, see
        `Facebook Audience API <https://developers.facebook.com/docs/marketing-api/audiences-api>`_,
        under "Hashing and Normalization for Multi-Key".

        Note that you shouldn't have to do normalization on your data, as long as it's
        reasonably close to what FB expects. Eg. It will convert "Male" to "m", and " JoSH"
        to "josh".

        FB will attempt to match the data to users in their system. You won't be able to find out
        which users were matched. But if you provide enough data, FB will tell you roughly how many
        of them were matched. (You can find the custom audience in your business account at
        https://business.facebook.com).

        Note that because FB's matching is so opaque, it will hide lots of data issues. Eg. if you
        use "United States" instead of "US" for the "country" field, the API will appear to accept
        it, when in reality it is probably ignoring that field. So read the docs if you're worried.

        `Args:`
            audience_id: str
                The ID of the custom audience to delete.
            users_table: obj
                Parsons table

        """ # noqa: E501,E261

        logger.info(f"Adding custom audience users from provided table with "
                    f"{users_table.num_rows} rows")

        match_table = FacebookAds.get_match_table_for_users_table(users_table)
        if not match_table.columns:
            raise KeyError("No valid columns found for audience matching. "
                           "See FacebookAds.KeyMatchMap for supported columns")

        num_rows = match_table.num_rows
        logger.info(f"Found {num_rows} rows with valid FB matching keys")
        logger.info(f"Using FB matching keys: {match_table.columns}")

        (schema, data) = FacebookAds._get_match_schema_and_data(match_table)

        # Use the FB API to add users, respecting the limit per API call.
        # Process and upload batches in parallel, to improve performance.

        batch_size = MAX_FB_AUDIENCE_API_USERS

        parallel_jobs = (
            delayed(FacebookAds._add_batch_to_custom_audience)(
                self.app_id, self.app_secret, self.access_token, audience_id, schema,
                data[i:i+batch_size], i, num_rows
            )
            for i in range(0, len(data), batch_size)
        )

        n_jobs = os.environ.get('PARSONS_NUM_PARALLEL_JOBS', 4)
        Parallel(n_jobs=n_jobs)(parallel_jobs)



import json
import logging
import requests
from parsons.utilities import check_env

logger = logging.getLogger(__name__)


class ActionKit(object):
    """
    Instatiate the ActionKit class

    `Args:`
        domain: str
            The ActionKit domain (e.g. ``myorg.actionkit.com``) Not required if
            ``ACTION_KIT_DOMAIN`` env variable set.
        username: str
            The authorized ActionKit username. Not required if ``ACTION_KIT_USERNAME`` env
            variable set.
        password: str
            The authorized ActionKit user password. Not required if ``ACTION_KIT_USERNAME``
            env variable set.
    """

    def __init__(self, domain=None, username=None, password=None):

        self.domain = check_env.check('ACTION_KIT_DOMAIN', domain)
        self.username = check_env.check('ACTION_KIT_USERNAME', username)
        self.password = check_env.check('ACTION_KIT_PASSWORD', password)
        self.conn = self._conn()

    def _conn(self):

        client = requests.Session()
        client.auth = (self.username, self.password)
        client.headers.update({
            'content-type': 'application/json',
            'accepts': 'application/json'
        })
        return client

    def _base_endpoint(self, endpoint, entity_id=None):
        # Create the base endpoint URL

        url = f'https://{self.domain}/rest/v1/{endpoint}/'

        if entity_id:
            return url + f'{entity_id}/'
        return url

    def _base_get(self, endpoint, entity_id=None, exception_message=None):
        # Make a general get request to ActionKit

        resp = self.conn.get(self._base_endpoint(endpoint, entity_id))

        if exception_message and resp.status_code == 404:
            raise Exception(exception_message)

        return resp.json()

    def _base_post(self, endpoint, exception_message, **kwargs):
        # Make a general post request to ActionKit

        resp = self.conn.post(self._base_endpoint(endpoint), data=json.dumps(kwargs))

        if resp.status_code != 201:
            raise Exception(exception_message)

        if 'headers' in resp.__dict__:
            return resp.__dict__['headers']['Location']

    def get_user(self, user_id):
        """
        Get a user.

        `Args:`
            user_id: int
                The user id of the record to get.
        `Returns`:
            User json object
        """

        return self._base_get(endpoint='user', entity_id=user_id,
                              exception_message='User not found')

    def get_user_fields(self):
        """
        Get list of valid user fields that can be passed with the
        :meth:`ActionKit.create_user` method.

        `Returns`:
            List of user fields
        """

        resp = self._base_get(endpoint='user/schema')

        return list(resp['fields'].keys())

    def create_user(self, email, **kwargs):
        """
        Create a user.

        `Args:`
            email: str
                Email for the user
            user_dict: dict
                Optional; Additional user fields
        `Returns:`
            User json object
        """

        return self._base_post(endpoint='user', exception_message='Could not create user',
                               email=email, **kwargs)

    def update_user(self, user_id, **kwargs):
        """
        Update a user.

        `Args:`
            user_id: int
                The user id of the person to update
            user_dict: dict
                A dictionary of fields to update for the user.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(self._base_endpoint('user', user_id), data=json.dumps(kwargs))
        logger.info(f'{resp.status_code}: {user_id}')

    def delete_user(self, user_id):
        """
        Delete a user.

        `Args:`
            user_id: int
                The user id of the person to delete
        `Returns:`
            ``None``
        """

        resp = self.conn.delete(self._base_endpoint('user', user_id))
        logger.info(f'{resp.status_code}: {user_id}')

    def get_campaign(self, campaign_id):
        """
        Get a campaign.

        `Args:`
            campaign_id: int
                The campaign id of the record.
        `Returns`:
            Campaign json object
        """

        return self._base_get(endpoint='campaign', entity_id=campaign_id,
                              exception_message='Campaign not found')

    def get_campaign_fields(self):
        """
        Get list of valid campaign fields that can be passed with the
        :meth:`ActionKit.create_campaign` and :meth:`ActionKit.update_campaign` methods.

        `Returns`:
            List of campaign fields
        """

        resp = self._base_get(endpoint='campaign/schema')
        return list(resp['fields'].keys())

    def create_campaign(self, name, **kwargs):
        """
        Create a campaign.

        `Args:`
            name: str
                The name of the campaign to create
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='campaign', exception_message='Could not create campaign',
                               name=name, **kwargs)

    def get_event_create_page(self, event_create_page_id):
        """
        Get a event create page.

        `Args:`
            event_create_page_id: int
                The event create page id of the record to get.
        `Returns`:
            Event create page json object
        """

        return self._base_get(endpoint='eventcreatepage', entity_id=event_create_page_id,
                              exception_message='Event create page not found')

    def get_event_create_page_fields(self):
        """
        Get list of event create page fields that can be passed with the
        :meth:`ActionKit.create_event_create_page`.

        `Returns`:
            List of event create page fields
        """

        resp = self._base_get(endpoint='eventcreatepage/schema')
        return list(resp['fields'].keys())

    def create_event_create_page(self, name, campaign_id, title, **kwargs):
        """
        Add an event page to a campaign.

        `Args:`
            campaign_id: int
                The campaign to assoicate page with
            name: str
                The name of the page to create
            title: str
                The title of the page to create
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='eventcreatepage',
                               exception_message='Could not create event create page',
                               campaign=f'/rest/v1/campaign/{campaign_id}/',
                               name=name,
                               title=title,
                               **kwargs)

    def get_event_create_form(self, event_create_form_id):
        """
        Get a event create form.

        `Args:`
            event_create_form_id: int
                The event create form id of the record to get.
        `Returns`:
            Event create form json object
        """

        return self._base_get(endpoint='eventcreateform', entity_id=event_create_form_id,
                              exception_message='Event create page not found')

    def get_event_create_form_fields(self):
        """
        Get list of valid event create form fields that can be passed with the
        :meth:`ActionKit.create_event_create_form` method.

        `Returns`:
            List of event create form fields
        """

        resp = self._base_get(endpoint='eventcreateform/schema')
        return list(resp['fields'].keys())

    def create_event_create_form(self, page_id, thank_you_text, **kwargs):
        """
        Create a event create form.

        `Args:`
            page_id: int
                The page to associate the form with
            thank_you_text: str
                Free form thank you text
        `Returns:`
            API location of new resource
        """

        return self._base_post(endpoint='eventcreateform',
                               exception_message='Could not event create form',
                               page=f'/rest/v1/eventcreatepage/{page_id}/',
                               thank_you_text=thank_you_text,
                               **kwargs)

    def get_event_signup_page(self, event_signup_page_id):
        """
        Get event signup page.

        `Args:`
            event_signup_page_id: int
                The event signup page id of the record to get.
        `Returns`:
            Event signup page json object
        """

        return self._base_get(endpoint='eventsignuppage', entity_id=event_signup_page_id,
                              exception_message='User page signup page not found')

    def get_event_signup_page_fields(self):
        """
        Get list of valid event signup page fields that can be passed with the
        :meth:`ActionKit.create_event_signup_page` method.

        `Returns`:
            List of event signup page fields
        """

        resp = self._base_get(endpoint='eventsignuppage/schema')
        return list(resp['fields'].keys())

    def create_event_signup_page(self, name, campaign_id, title, **kwargs):
        """
        Add an event signup page to a campaign.

        `Args:`
            campaign_id: int
                The campaign to assoicate page with
            name: str
                The name of the page to create
            title: str
                The title of the page to create
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='eventsignuppage',
                               exception_message='Could not create signup page',
                               campaign=f'/rest/v1/campaign/{campaign_id}/',
                               name=name,
                               title=title,
                               **kwargs)

    def get_event_signup_form(self, event_signup_form_id):
        """
        Get a user.

        `Args:`
            event_signup_form_id: str
                The event signup form id of the record to get.
        `Returns`:
            Event signup form json object
        """

        return self._base_get(endpoint='eventsignupform', entity_id=event_signup_form_id,
                              exception_message='User page signup form not found')

    def get_event_signup_form_fields(self):
        """
        Get list of valid event signup form fields that can be passed with the
        :meth:`ActionKit.create_event_signup_form` method.

        `Returns`:
            List of event signup form fields
        """

        resp = self._base_get(endpoint='eventsignupform/schema')
        return list(resp['fields'].keys())

    def create_event_signup_form(self, page_id, thank_you_text, **kwargs):
        """
        Create a event signup form.

        `Args:`
            page_id: int
                The page to associate the form with
            thank_you_text: str
                Free form thank you text
        `Returns:`
            API location of new resource
        """

        return self._base_post(endpoint='eventsignupform',
                               exception_message='Could not event create signup form',
                               page=f'/rest/v1/page/{page_id}/',
                               thank_you_text=thank_you_text,
                               **kwargs)

    def get_page_followup(self, page_followup_id):
        """
        Get a page followup.

        `Args:`
            page_followup_id: int
                The user id of the record to get.
        `Returns`:
            Page followup json object
        """

        return self._base_get(endpoint='pagefollowup', entity_id=page_followup_id,
                              exception_message='Page followup not found')

    def get_page_followup_fields(self):
        """
        Get list of valid page followup fields that can be passed with the
        :meth:`ActionKit.create_page_followup` method.

        `Returns`:
            List of page followup fields
        """

        resp = self._base_get(endpoint='pagefollowup/schema')
        return list(resp['fields'].keys())

    def create_page_followup(self, signup_page_id, url, **kwargs):
        """
        Add a page followup.

        `Args:`
            signup_page_id: int
                The signup page to associate the followup page with
            url: str
                URL of the folloup page
        `Returns`:
            API location of new resource
        """

        return self._base_post(endpoint='pagefollowup',
                               exception_message='Could not create page followup',
                               page=f'/rest/v1/eventsignuppage/{signup_page_id}/',
                               url=url,
                               **kwargs)

    # TO DO
    # def upload_file(self, page, file_name):
    #     """
    #     Further Testing - Upload file to ActionKit
    #
    #     `Args:`
    #         page: str
    #             Path for the
    #         file_name: str
    #             The path for the file_name
    #     """
    #     with open(file_name, 'rb') as local_file:
    #         resp = self.conn.post(
    #             self._base_endpoint('upload'),
    #             files={'upload': local_file},
    #             data={
    #                 'page': page,
    #                 'autocreate_user_fields': 'true'
    #             }
    #         )
    #
    #     return resp.status_code



from airtable import Airtable as AT
from parsons import Table
from parsons.utilities import check_env
import logging


logger = logging.getLogger(__name__)


class Airtable(object):
    """
    `Args:`
        base_key: str
            The key of the Airtable base that you will interact with.
        table_name: str
            The name of the table in the base. The table name is the equivilant of the sheet name
            in Excel or GoogleDocs.
        api_key: str
            The Airtable provided api key. Not required if ``AIRTABLE_API_KEY`` env variable set.
    """

    def __init__(self, base_key, table_name, api_key=None):

        self.api_key = check_env.check('AIRTABLE_API_KEY', api_key)
        self.at = AT(base_key, table_name, self.api_key)

    def get_record(self, record_id):
        """
        Returns a single record.

        `Args:`
            record_id: str
                The Airtable record id
        `Returns:`
            A dictionary of the record
        """

        return self.at.get(record_id)

    def get_records(self, fields=None, max_records=None, view=None, formula=None, sort=None):
        """
        `Args:`
            fields: str or lst
                Only return specified column or list of columns. The column name is
                case sensitive
            max_records: int
                The maximum total number of records that will be returned.
            view: str
                If set, only the records in that view will be returned. The records will be
                sorted according to the order of the view.
            formula: str
                The formula will be evaluated for each record, and if the result
                is not 0, false, "", NaN, [], or #Error! the record will be included
                in the response.

                If combined with view, only records in that view which satisfy the
                formula will be returned. For example, to only include records where
                ``COLUMN_A`` isn't empty, pass in: ``"NOT({COLUMN_A}='')"``

                For more information see
                `Airtable Docs on formulas. <https://airtable.com/api>`_

                Usage - Text Column is not empty:

                ``airtable.get_all(formula="NOT({COLUMN_A}='')")``

                Usage - Text Column contains:

                ``airtable.get_all(formula="FIND('SomeSubText', {COLUMN_STR})=1")``

            sort: str or lst
                Specifies how the records will be ordered. If you set the view parameter, the
                returned records in that view will be sorted by these fields. If sorting by
                multiple columns, column names can be passed as a list. Sorting Direction is
                ascending by default, but can be reversed by prefixing the column name with a minus
                sign -.

                Example usage:
                ``airtable.get(sort=['ColumnA', '-ColumnB'])``

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        # Raises an error if sort is None type. Thus, only adding if populated.
        kwargs = {'fields': fields, 'max_records': max_records, 'formula': formula}
        if sort:
            kwargs['sort'] = sort

        tbl = Table(self.at.get_all(**kwargs))
        return tbl.unpack_dict(column='fields', prepend=False)

    def insert_record(self, row):
        """
        Insert a single record into an Airtable.

        `Args:`
            row: dict
                Fields to insert. Must be dictionary with Column names as Key.
            typecast: boolean
                Automatic data conversion from string values.
        `Returns:`
            Dictionary of inserted row

        """

        resp = self.at.insert(row)
        logger.info('Record inserted')
        return resp

    def insert_records(self, table):
        """
        Insert multiple records into an Airtable. The columns in your Parsons table must
        exist in the Airtable. The method will attempt to map based on column name, so the
        order of the columns is irrelevant.

        `Args:`
            table: A Parsons Table
                Insert a Parsons table
            typecast: boolean
                Automatic data conversion from string values.
        `Returns:`
            List of dictionaries of inserted rows
        """

        resp = self.at.batch_insert(table)
        logger.info(f'{table.num_rows} records inserted.')
        return resp

    def update_record(self, record_id, fields, typecast=False):
        """
        Updates a record by its record id. Only Fields passed are updated, the rest are left as
        is.

        `Args:`
            record_id: str
                The Airtable record id
            fields: dict
                Fields to insert. Must be dictionary with Column names as Key.
            typecast: boolean
                Automatic data conversion from string values.
        `Returns:`
            ``None``
        """

        resp = self.at.update(record_id, fields, typecast=typecast)
        logger.info(f'{record_id} updated')
        return resp



from parsons.targetsmart.targetsmart_automation import TargetSmartAutomation
from parsons import SFTP
import unittest
from test.utils import mark_live_test

class TestTargetSmartAutomation(unittest.TestCase):

    def setUp(self):

        self.ts = TargetSmartAutomation()
        self.job_name = 'a-test-job'
        self.sftp = SFTP(self.ts.sftp_host,
                         os.environ['TS_SFTP_USERNAME'],
                         os.environ['TS_SFTP_PASSWORD'],
                         self.ts.sftp_port)
        self.test_xml = 'test/test_ts/job_config.xml'

    def tearDown(self):

        # Clean up the files were put on the SFTP
        self.ts.remove_files(self.job_name)

    @mark_live_test
    def test_create_job_xml(self):

        # Assert that job xml creates the file correctly
        job_xml = self.ts.create_job_xml('job_type',
                                         'match_job',
                                         ['test@gmail.com', 'test2@gmail.com'])
        with open(self.test_xml, 'r') as xml:
            test_xml = xml.read()
        with open(job_xml, 'r') as xml:
            real_xml = xml.read()
        self.assertEqual(test_xml, real_xml)

    @mark_live_test
    def test_config_status(self):

        # Find good configuration
        self.sftp.put_file(self.test_xml, f'{self.ts.sftp_dir}/{self.job_name}.job.xml.good')
        self.assertTrue(self.ts.config_status(self.job_name))
        self.ts.remove_files(self.job_name)

        # Find bad configuration
        self.sftp.put_file(self.test_xml, f'{self.ts.sftp_dir}/{self.job_name}.job.xml.bad')
        self.assertRaises(ValueError, self.ts.config_status, self.job_name)

    @mark_live_test
    def test_match_status(self):

        # Find good configuration
        good_match = 'test/test_ts/match_good.xml'
        self.sftp.put_file(good_match, f'{self.ts.sftp_dir}/{self.job_name}.finish.xml')
        self.assertTrue(self.ts.match_status(self.job_name))
        self.ts.remove_files(self.job_name)

        # Find bad configuration
        bad_match = 'test/test_ts/match_bad.xml'
        self.sftp.put_file(bad_match, f'{self.ts.sftp_dir}/{self.job_name}.finish.xml')
        self.assertRaises(ValueError, self.ts.match_status, self.job_name)

    @mark_live_test
    def test_remove_files(self):

        # Add a file
        self.sftp.put_file(self.test_xml, f'{self.ts.sftp_dir}/{self.job_name}.txt')

        # Remove files
        self.ts.remove_files(self.job_name)

        # Check that file is not there
        dir_list = self.sftp.list_directory(f'{self.ts.sftp_dir}/')
        self.assertNotIn(f'{self.job_name}.txt', dir_list)


"""Mobile Commons Connector."""

import requests
import os
import xmltodict
import json
import logging
from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class MobileCommonsConnector(object):
    """Mobile Commons Connector Class."""

    def __init__(self, username=None, password=None, company=None,
                 uri='https://secure.mcommons.com/api/'):
        """Initialize the MobileCommonsConnector class.

        `Args:`
            username: str
                The username for a mobile commons account with API permissions.
                If it's not passed in, it will attempt to get it from the
                environement, MC_USER.
            password: str
                The password associated with the username. If it's not passed
                in, it will attempt to get it from the environement, MC_PASS.
            company: str
                Specify which account/company you want to use for this session.
                If you leave it off, your default account will be used.
            uri: str
                Base uri to make api calls.
        """
        if username is None:

            try:
                username = os.environ['MC_USER']
            except KeyError:
                raise KeyError('No Mobile Commons username found. Store as '
                               'MC_USER environment variable or pass as an '
                               'argument.')

        if password is None:

            try:
                password = os.environ['MC_PASS']
            except KeyError:
                raise KeyError('No Mobile Commons password found. Store as '
                               'MC_PASS environment variable or pass as an '
                               'argument.')

        self.uri = uri
        self.username = username
        self.password = password
        self.company = company

    def request(self, url, req_type='GET', post_data=None, args=None,
                raw=False, resp_type='json'):
        """Send a request - internal."""
        auth = (self.username, self.password)

        if self.company:
            if not args:
                args = {}
            args['company'] = self.company

        if req_type == 'GET':
            r = requests.get(url, auth=auth, params=args)

        if req_type == 'POST':
            r = requests.post(url, auth=auth, json=post_data, params=args)

        if req_type == 'PATCH':
            r = requests.patch(url, auth=auth, json=post_data, params=args)

        if req_type == 'PUT':
            r = requests.put(url, auth=auth, json=post_data, params=args)

        if req_type == 'DELETE':
            r = requests.delete(url, auth=auth, params=args)

        # TODO: Parse all of the errors and stuff.

        if raw:

            return r

        else:
            if resp_type == 'xml':
                result = json.loads(json.dumps(xmltodict.parse(r.text)))
                return self.clean_dict(result)
            else:
                return r.json()

    def request_paginate(self, url, rsrc, req_type='GET', post_data=None,
                         args=None, resp_type='json'):
        """Paginate through request - internal."""
        items = []
        next_page = True
        rsrc_s = rsrc[:-1]

        counter = None
        count = 0

        while next_page:

            i = self.request(url, req_type=req_type, post_data=post_data,
                             args=args, resp_type=resp_type)

            if not counter:
                if i['response'][rsrc].get('num'):
                    counter = 'num'
                else:
                    counter = 'page_count'

            if int(i['response'][rsrc][counter]) == 0:
                next_page = False
                break

            items.extend(i['response'][rsrc][rsrc_s])

            args['page'] = int(i['response'][rsrc]['page']) + 1

            count += int(i['response'][rsrc][counter])

            if args.get('limit') and count >= args['limit']:
                next_page = False
                break

        return items

    def clean_dict(self, d, rem='@'):
        """Recursively remove a string from the keys if a dictself.

        `Args:`
            d: list or dict
                The list or dict to be cleaned.
            rem: str
                The string to remove from the key.
        `Returns:`
            list or dict depending on the input
        """
        if isinstance(d, list):
            newl = []
            for e in d:
                if isinstance(e, dict):
                    newl.append(self.clean_dict(e))
                else:
                    newl.append(e)
            return newl

        else:
            new = {}

            for k, v in d.items():
                if isinstance(v, dict):
                    v = self.clean_dict(v)

                if isinstance(v, list):
                    newl = []
                    for e in v:
                        if isinstance(e, dict):
                            newl.append(self.clean_dict(e))
                        else:
                            newl.append(e)
                    new[k.replace(rem, '')] = newl
                    continue

                new[k.replace(rem, '')] = v
            return new

    def output(self, list_obj):
        """Return an object as a Table.

        `Args:`
            list_obj: dict or list
                The list or dict to be output.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        if isinstance(list_obj, dict):
            list_obj = [list_obj]

        return Table(list_obj)



from parsons import Redshift
from parsons import S3
from parsons.etl.table import Table
from test.utils import assert_matching_tables
import unittest
import os
import re
import warnings
import datetime
from test.utils import validate_list


# The name of the schema and will be temporarily created for the tests
TEMP_SCHEMA = 'parsons_test'

# These tests do not interact with the Redshift Database directly, and don't need real credentials


class TestRedshift(unittest.TestCase):

    def setUp(self):

        self.rs = Redshift(username='test', password='test', host='test', db='test', port=123)

        self.tbl = Table([['ID', 'Name'],
                          [1, 'Jim'],
                          [2, 'John'],
                          [3, 'Sarah']])

        self.mapping = self.rs.generate_data_types(self.tbl)

    def tearDown(self):

        pass

    def test_split_full_table_name(self):
        schema, table = Redshift.split_full_table_name('some_schema.some_table')
        self.assertEqual(schema, 'some_schema')
        self.assertEqual(table, 'some_table')

        # When missing the schema
        schema, table = Redshift.split_full_table_name('some_table')
        self.assertEqual(schema, 'public')
        self.assertEqual(table, 'some_table')

        # When there are too many parts
        self.assertRaises(ValueError, Redshift.split_full_table_name, 'a.b.c')

    def test_combine_schema_and_table_name(self):
        full_table_name = Redshift.combine_schema_and_table_name('some_schema', 'some_table')
        self.assertEqual(full_table_name, 'some_schema.some_table')

    def test_data_type(self):

        # Test smallint
        self.assertEqual(self.rs.data_type(1, ''), 'smallint')
        # Test int
        self.assertEqual(self.rs.data_type(32769, ''), 'int')
        # Test bigint
        self.assertEqual(self.rs.data_type(2147483648, ''), 'bigint')
        # Test varchar that looks like an int
        self.assertEqual(self.rs.data_type('00001', ''), 'varchar')
        # Test a float as a decimal
        self.assertEqual(self.rs.data_type(5.001, ''), 'decimal')
        # Test varchar
        self.assertEqual(self.rs.data_type('word', ''), 'varchar')
        # Test int with underscore
        self.assertEqual(self.rs.data_type('1_2', ''), 'varchar')
        # Test int with leading zero
        self.assertEqual(self.rs.data_type('01', ''), 'varchar')

    def test_generate_data_types(self):

        # Test correct header labels
        self.assertEqual(self.mapping['headers'], ['ID', 'Name'])
        # Test correct data types
        self.assertEqual(self.mapping['type_list'], ['smallint', 'varchar'])
        # Test correct lengths
        self.assertEqual(self.mapping['longest'], [1, 5])

    def test_vc_padding(self):

        # Test padding calculated correctly
        self.assertEqual(self.rs.vc_padding(self.mapping, .2), [1, 6])

    def test_vc_max(self):

        # Test max sets it to the max
        self.assertEqual(self.rs.vc_max(self.mapping, ['Name']), [1, 65535])

        # Test raises when can't find column
        # To Do

    def test_vc_validate(self):

        # Test that a column with a width of 0 is set to 1
        self.mapping['longest'][0] = 0
        self.mapping = self.rs.vc_validate(self.mapping)
        self.assertEqual(self.mapping, [1, 5])

    def test_create_sql(self):

        # Test the the statement is expected
        sql = self.rs.create_sql('tmc.test', self.mapping, distkey='ID')
        exp_sql = "create table tmc.test (\n  id smallint,\n  name varchar(5)) \ndistkey(ID) ;"
        self.assertEqual(sql, exp_sql)

    def test_column_validate(self):

        bad_cols = ['', 'SELECT', 'asdfjkasjdfklasjdfklajskdfljaskldfjaklsdfjlaksdfjklasjdfklasjdkfljaskldfljkasjdkfasjlkdfjklasdfjklakjsfasjkdfljaslkdfjklasdfjklasjkldfakljsdfjalsdkfjklasjdfklasjdfklasdkljf'] # noqa: E501
        fixed_cols = ['col_0', 'col_1', 'asdfjkasjdfklasjdfklajskdfljaskldfjaklsdfjlaksdfjklasjdfklasjdkfljaskldfljkasjdkfasjlkdfjklasdfjklakjsfasjkdfljaslkdfjkl'] # noqa: E501
        self.assertEqual(self.rs.column_name_validate(bad_cols), fixed_cols)

    def test_create_statement(self):

        # Assert that copy statement is expected
        sql = self.rs.create_statement(self.tbl, 'tmc.test', distkey='ID')
        exp_sql = """create table tmc.test (\n  "id" smallint,\n  "name" varchar(5)) \ndistkey(ID) ;"""  # noqa: E501
        self.assertEqual(sql, exp_sql)

        # Assert that an error is raised by an empty table
        empty_table = Table([['Col_1', 'Col_2']])
        self.assertRaises(ValueError, self.rs.create_statement, empty_table, 'tmc.test')

    def test_get_creds_kwargs(self):

        # Test passing kwargs
        creds = self.rs.get_creds('kwarg_key', 'kwarg_secret_key')
        expected = """credentials 'aws_access_key_id=kwarg_key;aws_secret_access_key=kwarg_secret_key'\n"""  # noqa: E501
        self.assertEqual(creds, expected)

        # Test grabbing from environmental variables
        prior_aws_access_key_id = os.environ.get('AWS_ACCESS_KEY_ID', '')
        prior_aws_secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY', '')
        os.environ['AWS_ACCESS_KEY_ID'] = 'env_key'
        os.environ['AWS_SECRET_ACCESS_KEY'] = 'env_secret_key'
        creds = self.rs.get_creds(None, None)
        expected = """credentials 'aws_access_key_id=env_key;aws_secret_access_key=env_secret_key'\n"""  # noqa: E501
        self.assertEqual(creds, expected)

        # Reset env vars
        os.environ['AWS_ACCESS_KEY_ID'] = prior_aws_access_key_id
        os.environ['AWS_SECRET_ACCESS_KEY'] = prior_aws_secret_access_key

    def scrub_copy_tokens(self, s):

        s = re.sub('=.+;', '=*HIDDEN*;', s)
        s = re.sub('aws_secret_access_key=.+\'',
                   'aws_secret_access_key=*HIDDEN*\'', s)
        return s

    def test_copy_statement(self):

        sql = self.rs.copy_statement('test_schema.test', 'buck', 'file.csv',
            aws_access_key_id='abc123', aws_secret_access_key='abc123')

        # Scrub the keys
        sql = re.sub(r'id=.+;', '*id=HIDDEN*;', re.sub(r"key=.+'", "key=*HIDDEN*'", sql))

        expected_options = ['statupdate', 'compupdate', 'ignoreheader 1', 'acceptanydate',
                            "dateformat 'auto'", "timeformat 'auto'", "csv delimiter ','",
                            "copy test_schema.test \nfrom 's3://buck/file.csv'",
                            "'aws_access_key_*id=HIDDEN*;aws_secret_access_key=*HIDDEN*'",
                            'emptyasnull', 'blanksasnull', 'acceptinvchars']

        # Check that all of the expected options are there:
        [self.assertNotEqual(sql.find(o), -1) for o in expected_options]

    def test_copy_statement_columns(self):

        cols = ['a', 'b', 'c']

        sql = self.rs.copy_statement('test_schema.test', 'buck', 'file.csv',
            aws_access_key_id='abc123', aws_secret_access_key='abc123', specifycols=cols)

        # Scrub the keys
        sql = re.sub(r'id=.+;', '*id=HIDDEN*;', re.sub(r"key=.+'", "key=*HIDDEN*'", sql))

        print(sql)

        expected_options = ['statupdate', 'compupdate', 'ignoreheader 1', 'acceptanydate',
                            "dateformat 'auto'", "timeformat 'auto'", "csv delimiter ','",
                            "copy test_schema.test(a, b, c) \nfrom 's3://buck/file.csv'",
                            "'aws_access_key_*id=HIDDEN*;aws_secret_access_key=*HIDDEN*'",
                            'emptyasnull', 'blanksasnull', 'acceptinvchars']
        for o in expected_options:
            print(o, sql.find(o))

        # Check that all of the expected options are there:
        [self.assertNotEqual(sql.find(o), -1) for o in expected_options]

# These tests interact directly with the Redshift database


@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestRedshiftDB(unittest.TestCase):

    def setUp(self):

        self.temp_schema = TEMP_SCHEMA

        self.rs = Redshift()

        self.tbl = Table([['ID', 'Name'],
                          [1, 'Jim'],
                          [2, 'John'],
                          [3, 'Sarah']])

        # Create a schema, create a table, create a view
        setup_sql = f"""
                    drop schema if exists {self.temp_schema} cascade;
                    create schema {self.temp_schema};
                    """

        other_sql = f"""
                    create table {self.temp_schema}.test (id smallint,name varchar(5));
                    create view {self.temp_schema}.test_view as (select * from {self.temp_schema}.test);
                    """ # noqa: E501

        self.rs.query(setup_sql)

        self.rs.query(other_sql)

        self.s3 = S3()

        self.temp_s3_bucket = os.environ['S3_TEMP_BUCKET']
        self.temp_s3_prefix = 'test/'

    def tearDown(self):

        # Drop the view, the table and the schema
        teardown_sql = f"""
                       drop schema if exists {self.temp_schema} cascade;
                       """
        self.rs.query(teardown_sql)

        # Remove all test objects from S3
        for key in self.s3.list_keys(self.temp_s3_bucket, self.temp_s3_prefix):
            self.s3.remove_file(self.temp_s3_bucket, key)

    def test_query(self):

        # Check that query sending back expected result
        r = self.rs.query('select 1')
        self.assertEqual(r[0]['?column?'], 1)

    def test_query_with_parameters(self):
        table_name = f"{self.temp_schema}.test"
        self.tbl.to_redshift(table_name, if_exists='append')

        sql = f"select * from {table_name} where name = %s"
        name = 'Sarah'
        r = self.rs.query(sql, parameters=[name])
        self.assertEqual(r[0]['name'], name)

        sql = f"select * from {table_name} where name in (%s, %s)"
        names = ['Sarah', 'John']
        r = self.rs.query(sql, parameters=names)
        self.assertEqual(r.num_rows, 2)

    def test_schema_exists(self):
        self.assertTrue(self.rs.schema_exists(self.temp_schema))
        self.assertFalse(self.rs.schema_exists('nonsense'))

    def test_table_exists(self):

        # Check if table_exists finds a table that exists
        self.assertTrue(self.rs.table_exists(f'{self.temp_schema}.test'))

        # Check if table_exists is case insensitive
        self.assertTrue(self.rs.table_exists(f'{self.temp_schema.upper()}.TEST'))

        # Check if table_exists doesn't find a table that doesn't exists
        self.assertFalse(self.rs.table_exists(f'{self.temp_schema}.test_fake'))

        # Check if table_exists finds a table that exists
        self.assertTrue(self.rs.table_exists(f'{self.temp_schema}.test_view'))

        # Check if table_exists doesn't find a view that doesn't exists
        self.assertFalse(self.rs.table_exists(f'{self.temp_schema}.test_view_fake'))

        # Check that the view kwarg works
        self.assertFalse(self.rs.table_exists(f'{self.temp_schema}.test_view', view=False))

    def test_temp_s3_create(self):

        key = self.rs.temp_s3_copy(self.tbl)

        # Test that you can get the object
        self.s3.get_file(self.temp_s3_bucket, key)

        # Try to delete the object
        self.rs.temp_s3_delete(key)

    def test_copy_s3(self):

        # To Do
        pass

    def test_copy(self):

        # Copy a table
        self.rs.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')

        # Test that file exists
        r = self.rs.query(f"select * from {self.temp_schema}.test_copy where name='Jim'")
        self.assertEqual(r[0]['id'], 1)

        # Copy to the same table, to verify that the "truncate" flag works.
        self.rs.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='truncate')
        rows = self.rs.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(rows[0]['count'], 3)

        # Copy to the same table, to verify that the "drop" flag works.
        self.rs.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')

    def test_upsert(self):

        # Create a target table
        self.rs.copy(self.tbl, f'{self.temp_schema}.test_copy')

        # Run upsert
        upsert_tbl = Table([['id', 'name'], [1, 'Jane'], [5, 'Bob']])
        self.rs.upsert(upsert_tbl, f'{self.temp_schema}.test_copy', 'ID')

        # Make sure that it is the expected table
        expected_tbl = Table([['id', 'name'], [1, 'Jane'], [2, 'John'], [3, 'Sarah'], [5, 'Bob']])
        updated_tbl = self.rs.query(f'select * from {self.temp_schema}.test_copy order by id;')
        assert_matching_tables(expected_tbl, updated_tbl)

        # Try to run it with a bad primary key
        self.rs.query(f"INSERT INTO {self.temp_schema}.test_copy VALUES (1, 'Jim')")
        self.assertRaises(ValueError, self.rs.upsert, upsert_tbl, f'{self.temp_schema}.test_copy', 'ID')

    def test_unload(self):

        # Copy a table to Redshift
        self.rs.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')

        # Unload a table to S3
        self.rs.unload(f'select * from {self.temp_schema}.test_copy',
                       self.temp_s3_bucket,
                       'unload_test')

        # Check that files are there
        self.assertTrue(self.s3.key_exists(self.temp_s3_bucket, 'unload_test'))

    def test_to_from_redshift(self):

        # Test the parsons table methods
        table_name = f'{self.temp_schema}.test_copy'
        self.tbl.to_redshift(table_name, if_exists='drop')
        sql = f"SELECT * FROM {table_name} ORDER BY id"
        result_tbl = Table.from_redshift(sql)
        # Don't bother checking columns names, since those were tweaked en route to Redshift.
        assert_matching_tables(self.tbl, result_tbl, ignore_headers=True)

    def test_generate_manifest(self):

        # Add some tables to buckets
        self.tbl.to_s3_csv(self.temp_s3_bucket, f'{self.temp_s3_prefix}test_file_01.csv')
        self.tbl.to_s3_csv(self.temp_s3_bucket, f'{self.temp_s3_prefix}test_file_02.csv')
        self.tbl.to_s3_csv(self.temp_s3_bucket, f'{self.temp_s3_prefix}test_file_03.csv')
        self.tbl.to_s3_csv(self.temp_s3_bucket, f'{self.temp_s3_prefix}dont_include.csv')

        # Copy in a table to generate the headers and table
        self.rs.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')

        # Generate the manifest
        manifest_key = f'{self.temp_s3_prefix}test_manifest.json'
        manifest = self.rs.generate_manifest(self.temp_s3_bucket,
                                             prefix=f'{self.temp_s3_prefix}test_file',
                                             manifest_bucket=self.temp_s3_bucket,
                                             manifest_key=manifest_key)

        # Validate path formatted correctly
        valid_url = f's3://{self.temp_s3_bucket}/{self.temp_s3_prefix}test_file_01.csv'
        self.assertEqual(manifest['entries'][0]['url'], valid_url)

        # Validate that there are three files
        self.assertEqual(len(manifest['entries']), 3)

        # Validate that manifest saved to bucket
        keys = self.s3.list_keys(self.temp_s3_bucket,
            prefix=f'{self.temp_s3_prefix}test_manifest')
        self.assertTrue(manifest_key in keys)

    def test_move_table(self):

        # Run the method and check that new table created
        self.rs.move_table(f'{self.temp_schema}.test', f'{self.temp_schema}.test2')
        self.assertTrue(self.rs.table_exists(f'{self.temp_schema}.test2'))

        # Run the method again, but drop original
        self.rs.move_table(f'{self.temp_schema}.test2', f'{self.temp_schema}.test3',
                                drop_source_table=True)
        self.assertFalse(self.rs.table_exists(f'{self.temp_schema}.test2'))

    def test_get_tables(self):

        tbls_list = self.rs.get_tables(schema=self.temp_schema)
        exp = ['schemaname', 'tablename', 'tableowner', 'tablespace', 'hasindexes',
               'hasrules', 'hastriggers']

        self.assertTrue(validate_list(exp, tbls_list))

    def test_get_table_stats(self):

        tbls_list = self.rs.get_table_stats(schema=self.temp_schema)
        exp = ['database', 'schema', 'table_id', 'table', 'encoded', 'diststyle', 'sortkey1',
               'max_varchar', 'sortkey1_enc', 'sortkey_num', 'size', 'pct_used', 'empty',
               'unsorted', 'stats_off', 'tbl_rows', 'skew_sortkey1', 'skew_rows']

        # Having some issues testing that the filter is working correctly, as it
        # takes a little bit of time for a table to show in this table and is beating
        # the test suite. I feel confident that it works though.

        self.assertTrue(validate_list(exp, tbls_list))

    def test_get_views(self):
        # Assert that get_views returns filtered views

        # Assert that it works with schema filter
        views = self.rs.get_views(schema=self.temp_schema)
        expected_row = (self.temp_schema,
                        'test_view',
                        f'SELECT test.id, test.name FROM {self.temp_schema}.test;')
        self.assertEqual(views.data[0], expected_row)


    def test_get_queries(self):

        # Validate that columns match expected columns
        queries_list = self.rs.get_queries()
        exp = ['user', 'pid', 'xid', 'query', 'service_class', 'slot', 'start', 'state',
               'queue_sec', 'exec_sec', 'cpu_sec', 'read_mb', 'spill_mb', 'return_rows',
               'nl_rows', 'sql', 'alert']
        self.assertTrue(validate_list(exp, queries_list))

    def test_get_row_count(self):
        table_name = f'{self.temp_schema}.test_row_count'
        self.rs.copy(self.tbl, table_name, if_exists='drop')
        count = self.rs.get_row_count(table_name)
        self.assertEqual(count, 3)

    def test_rename_table(self):

        self.rs.rename_table(self.temp_schema + '.test', 'test2')

        # Test that renamed table exists
        self.assertTrue(self.rs.table_exists(self.temp_schema + '.test2'))

        # Test that old table name does not exist
        self.assertFalse(self.rs.table_exists(self.temp_schema + '.test'))

    def test_union_tables(self):

        # Copy in two tables
        self.rs.copy(self.tbl, f'{self.temp_schema}.union_base1', if_exists='drop')
        self.rs.copy(self.tbl, f'{self.temp_schema}.union_base2', if_exists='drop')

        # Union all the two tables and check row count
        self.rs.union_tables(f'{self.temp_schema}.union_all',
                             [f'{self.temp_schema}.union_base1', f'{self.temp_schema}.union_base2'])
        self.assertEqual(self.rs.query(f"select * from {self.temp_schema}.union_all").num_rows, 6)

        # Union the two tables and check row count
        self.rs.union_tables(f'{self.temp_schema}.union_test',
                             [f'{self.temp_schema}.union_base1', f'{self.temp_schema}.union_base2'],
                             union_all=False)
        self.assertEqual(self.rs.query(f"select * from {self.temp_schema}.union_test").num_rows, 3)

    def test_populate_table_from_query(self):
        # Populate the source table
        source_table = f'{self.temp_schema}.test_source'
        self.rs.copy(self.tbl, source_table, if_exists='drop')

        query = f"SELECT * FROM {source_table}"

        # Populate the table
        dest_table = f'{self.temp_schema}.test_dest'
        self.rs.populate_table_from_query(query, dest_table)

        # Verify
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 3)

        # Try with if_exists='truncate'
        self.rs.populate_table_from_query(query, dest_table, if_exists='truncate')
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 3)

        # Try with if_exists='drop', and a distkey
        self.rs.populate_table_from_query(query, dest_table, if_exists='drop', distkey='id')
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 3)

        # Try with if_exists='fail'
        self.assertRaises(ValueError,
            self.rs.populate_table_from_query, query, dest_table, if_exists='fail')

    def test_duplicate_table(self):
        # Populate the source table
        source_table = f'{self.temp_schema}.test_source'
        self.rs.copy(self.tbl, source_table, if_exists='drop')

        # Duplicate the table
        dest_table = f'{self.temp_schema}.test_dest'
        self.rs.duplicate_table(source_table, dest_table)

        # Verify
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 3)

        # Try with if_exists='truncate'
        self.rs.duplicate_table(source_table, dest_table, if_exists='truncate')
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 3)

        # Try with if_exists='drop'
        self.rs.duplicate_table(source_table, dest_table, if_exists='drop')
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 3)

        # Try with if_exists='append'
        self.rs.duplicate_table(source_table, dest_table, if_exists='append')
        rows = self.rs.query(f"select count(*) from {dest_table}")
        self.assertEqual(rows[0]['count'], 6)

        # Try with if_exists='fail'
        self.assertRaises(ValueError, self.rs.duplicate_table, source_table,
                          dest_table, if_exists='fail')

        # Try with invalid if_exists arg
        self.assertRaises(ValueError, self.rs.duplicate_table, source_table,
                          dest_table, if_exists='nonsense')

    def test_get_max_value(self):

        date_tbl = Table([['id', 'date_modified'], [1, '2020-01-01'], [2, '1900-01-01']])
        self.rs.copy(date_tbl, f'{self.temp_schema}.test_date')

        # Test return string
        self.assertEqual(self.rs.get_max_value(f'{self.temp_schema}.test_date', 'date_modified'),
                         '2020-01-01')

    def test_get_columns(self):
        cols = self.rs.get_columns(self.temp_schema, 'test')

        # id smallint,name varchar(5)
        expected_cols = {
            'id':   {'data_type': 'smallint', 'max_length': 16, 'is_nullable': True},
            'name': {'data_type': 'character varying', 'max_length': 5, 'is_nullable': True},
        }

        self.assertEqual(cols, expected_cols)

    def test_get_object_type(self):
        # Test a table
        expected_type_table = "table"
        actual_type_table = self.rs.get_object_type("pg_catalog.pg_class")

        self.assertEqual(expected_type_table, actual_type_table)

        # Test a view
        expected_type_view = "view"
        actual_type_view = self.rs.get_object_type("pg_catalog.pg_views")

        self.assertEqual(expected_type_view, actual_type_view)

        # Test a nonexisting table
        expected_type_fake = None
        actual_type_fake = self.rs.get_object_type("someschema.faketable")

        self.assertEqual(expected_type_fake, actual_type_fake)

    def test_is_view(self):

        self.assertTrue(self.rs.is_view("pg_catalog.pg_views"))

        self.assertFalse(self.rs.is_view("pg_catalog.pg_class"))

    def test_is_table(self):

        self.assertTrue(self.rs.is_table("pg_catalog.pg_class"))

        self.assertFalse(self.rs.is_table("pg_catalog.pg_views"))

    def test_get_table_definition(self):
        expected_table_def = (
            '--DROP TABLE pg_catalog.pg_amop;'
            '\nCREATE TABLE IF NOT EXISTS pg_catalog.pg_amop'
            '\n('
            '\n\tamopclaid OID NOT NULL  ENCODE RAW'
            '\n\t,amopsubtype OID NOT NULL  ENCODE RAW'
            '\n\t,amopstrategy SMALLINT NOT NULL  ENCODE RAW'
            '\n\t,amopreqcheck BOOLEAN NOT NULL  ENCODE RAW'
            '\n\t,amopopr OID NOT NULL  ENCODE RAW'
            '\n)\nDISTSTYLE EVEN\n;')
        actual_table_def = self.rs.get_table_definition("pg_catalog.pg_amop")

        self.assertEqual(expected_table_def, actual_table_def)

    def test_get_table_definitions(self):
        expected_table_defs = [{
            'tablename': 'pg_catalog.pg_amop',
            'ddl': '--DROP TABLE pg_catalog.pg_amop;'
                   '\nCREATE TABLE IF NOT EXISTS pg_catalog.pg_amop'
                   '\n('
                   '\n\tamopclaid OID NOT NULL  ENCODE RAW'
                   '\n\t,amopsubtype OID NOT NULL  ENCODE RAW'
                   '\n\t,amopstrategy SMALLINT NOT NULL  ENCODE RAW'
                   '\n\t,amopreqcheck BOOLEAN NOT NULL  ENCODE RAW'
                   '\n\t,amopopr OID NOT NULL  ENCODE RAW'
                   '\n)\nDISTSTYLE EVEN\n;'}, {
            'tablename': 'pg_catalog.pg_amproc',
            'ddl': '--DROP TABLE pg_catalog.pg_amproc;'
                   '\nCREATE TABLE IF NOT EXISTS pg_catalog.pg_amproc'
                   '\n('
                   '\n\tamopclaid OID NOT NULL  ENCODE RAW'
                   '\n\t,amprocsubtype OID NOT NULL  ENCODE RAW'
                   '\n\t,amprocnum SMALLINT NOT NULL  ENCODE RAW'
                   '\n\t,amproc REGPROC NOT NULL  ENCODE RAW'
                   '\n)'
                   '\nDISTSTYLE EVEN\n;'}]
        actual_table_defs = self.rs.get_table_definitions(table="pg_am%p%")

        self.assertEqual(expected_table_defs, actual_table_defs)

    def test_get_view_definition(self):
        expected_view_def = (
            '--DROP VIEW pg_catalog.pg_views;'
            '\nCREATE OR REPLACE VIEW pg_catalog.pg_views AS'
            '\n SELECT n.nspname AS schemaname'
            ', c.relname AS viewname'
            ', pg_get_userbyid(c.relowner) AS viewowner'
            ', pg_get_viewdef(c.oid) AS definition'
            '\n   FROM pg_class c'
            '\n   LEFT JOIN pg_namespace n ON n.oid = c.relnamespace'
            '\n  WHERE c.relkind = \'v\'::"char";')
        actual_view_def = self.rs.get_view_definition("pg_catalog.pg_views")

        self.assertEqual(expected_view_def, actual_view_def)

    def test_get_view_definitions(self):
        expected_view_defs = [{
            'viewname': 'pg_catalog.pg_class_info',
            'ddl': "--DROP VIEW pg_catalog.pg_class_info;"
            "\nCREATE OR REPLACE VIEW pg_catalog.pg_class_info AS"
            "\n SELECT pgc.oid AS reloid, pgc.relname, pgc.relnamespace, "
            "pgc.reltype, pgc.relowner, pgc.relam, pgc.relfilenode, "
            "pgc.reltablespace, pgc.relpages, pgc.reltuples, "
            "pgc.reltoastrelid, pgc.reltoastidxid, pgc.relhasindex, "
            "pgc.relisshared, pgc.relkind, pgc.relnatts, pgc.relexternid, "
            "pgc.relisreplicated, pgc.relispinned, pgc.reldiststyle, "
            "pgc.relprojbaseid, pgc.relchecks, pgc.reltriggers, pgc.relukeys, "
            "pgc.relfkeys, pgc.relrefs, pgc.relhasoids, pgc.relhaspkey, "
            "pgc.relhasrules, pgc.relhassubclass, pgc.relacl, "
            "pgce0.value::smallint AS releffectivediststyle, "
            "date_add('microsecond'::text, pgce1.value::bigint, "
            "'2000-01-01 00:00:00'::timestamp without time zone) AS "
            "relcreationtime"
            "\n   FROM pg_class pgc"
            "\n   LEFT JOIN pg_class_extended pgce0 "
            "ON pgc.oid = pgce0.reloid AND pgce0.colnum = 0"
            "\n   LEFT JOIN pg_class_extended pgce1 "
            "ON pgc.oid = pgce1.reloid AND pgce1.colnum = 1;"}]
        actual_view_def = self.rs.get_view_definitions(view="pg_c%")

        self.assertEqual(expected_view_defs, actual_view_def)


if __name__ == "__main__":
    unittest.main()



from parsons.utilities import check_env
import requests
from parsons import Table

URI = 'https://www.googleapis.com/civicinfo/v2/'


class GoogleCivic(object):
    """
    `Args:`
        api_key : str
            A valid Google api key. Not required if ``GOOGLE_CIVIC_API_KEY``
            env variable set.
    `Returns:`
        class
    """

    def __init__(self, api_key=None):

        self.api_key = check_env.check('GOOGLE_CIVIC_API_KEY', api_key)
        self.uri = URI

    def request(self, url, args=None):
        # Internal request method

        if not args:
            args = {}

        args['key'] = self.api_key

        r = requests.get(url, params=args)

        return r.json()

    def get_elections(self):
        """
        Get a collection of information about elections and voter information.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.uri + 'elections'

        return Table((self.request(url))['elections'])

    def _get_voter_info(self, election_id, address):
        # Internal method to call voter info end point. Portions of this are
        # parsed for other methods.

        url = self.uri + 'voterinfo'

        args = {'address': address, 'electionId': election_id}

        return self.request(url, args=args)

    def get_polling_location(self, election_id, address):
        """
        Get polling location information for a given address.

        `Args:`
            election_id: int
                A valid election id. Election ids can be found by running the
                :meth:`get_elections` method.
            address: str
                A valid US address in a single string.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self._get_voter_info(election_id, address)

        return r['pollingLocations']

    def get_polling_locations(self, election_id, table, address_field='address'):
        """
        Get polling location information for a table of addresses.

        `Args:`
            election_id: int
                A valid election id. Election ids can be found by running the
                :meth:`get_elections` method.
            address: str
                A valid US address in a single string.
            address_field: str
                The name of the column where the address is stored.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        polling_locations = []

        # Iterate through the rows of the table
        for row in table:
            loc = self.get_polling_location(election_id, row[address_field])
            # Insert original passed address
            loc[0]['passed_address'] = row[address_field]

            # Add to list of lists
            polling_locations.append(loc[0])

        # Unpack values
        tbl = Table(polling_locations)
        tbl.unpack_dict('address', prepend_value='polling')
        tbl.unpack_list('sources', replace=True)
        tbl.unpack_dict('sources_0', prepend_value='source')
        tbl.rename_column('polling_line1', 'polling_address')

        # Resort columns
        tbl.move_column('pollingHours', len(tbl.columns))
        tbl.move_column('notes', len(tbl.columns))
        tbl.move_column('polling_locationName', 1)
        tbl.move_column('polling_address', 2)

        return tbl



from parsons.etl.etl import ETL
from parsons.etl.tofrom import ToFrom
import petl
import logging


logger = logging.getLogger(__name__)

DIRECT_INDEX_WARNING_COUNT = 10


class Table(ETL, ToFrom):
    """
    Create a Parsons Table. Accepts one of the following:
    - A list of lists, with list[0] holding field names, and the other lists holding data
    - A list of dicts
    - A petl table

    `Args:`
        lst: list
            See above for accepted list formats
        source: str
            The original data source from which the data was pulled (optional)
        name: str
            The name of the table (optional)
    """

    def __init__(self, lst=[]):

        self.table = None

        lst_type = type(lst)

        if lst_type in [list, tuple]:

            # Check for empty list
            if not len(lst):
                self.table = petl.fromdicts([])
            else:
                row_type = type(lst[0])
                # Check for list of dicts
                if row_type == dict:
                    self.table = petl.fromdicts(lst)
                # Check for list of lists
                elif row_type in [list, tuple]:
                    self.table = petl.wrap(lst)

        else:
            # Create from a petl table
            self.table = lst

        if not self.is_valid_table():
            raise ValueError("Could not create Table")

        # Count how many times someone is indexing directly into this table, so we can warn
        # against inefficient usage.
        self._index_count = 0

    def __repr__(self):

        return repr(petl.dicts(self.table))

    def __iter__(self):

        return iter(petl.dicts(self.table))

    def __getitem__(self, index):

        self._index_count += 1
        if self._index_count >= DIRECT_INDEX_WARNING_COUNT:
            logger.warning("""
                You have indexed directly into this Table multiple times. This can be inefficient,
                as data transformations you've made will be computed _each time_ you index into the
                Table. If you are accessing many rows of data, consider switching to this style of
                iteration, which is much more efficient:
                `for row in table:`
                """)

        return petl.dicts(self.table)[index]

    @property
    def num_rows(self):
        """
        `Returns:`
            int
                Number of rows in the table
        """
        return petl.nrows(self.table)

    @property
    def data(self):
        """
        Returns an iterable object for iterating over the raw data rows as tuples
        (without field names)
        """
        return petl.data(self.table)

    @property
    def columns(self):
        """
        `Returns:`
            list
                List of the table's column names
        """
        return list(petl.header(self.table))

    @property
    def first(self):
        """
        Returns the first value in the table. Useful for database queries that only
        return a single value.
        """

        try:
            return self.data[0][0]

        # If first value is empty, return None
        except IndexError:
            return None

    def materialize(self):
        """
        "Materializes" a Table, meaning all data is loaded into memory and all pending
        transformations are applied.

        Use this if petl's lazy-loading behavior is causing you problems, eg. if you want to read
        data from a file immediately.
        """

        self.table = petl.wrap(petl.tupleoftuples(self.table))

    def is_valid_table(self):
        """
        Performs some simple checks on a Table. Specifically, verifies that we have a valid petl
        table within the Parsons Table.

        `Returns:`
            bool
        """

        if not self.table:
            return False

        try:
            self.columns
        except StopIteration:
            return False

        return True

    def empty_column(self, column):
        """
        Checks if a given column is empty. Returns ``True`` if empty and ``False``
        if not empty.

        `Args:`
            column: str
                The column name
        `Returns:`
            bool
        """

        if petl.nrows(petl.selectnotnone(self.table, column)) == 0:
            return True
        else:
            return False



"""NGPVAN Survey Questions Endpoints"""
from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class SurveyQuestions(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_survey_questions(self, statuses=['Active'], name=None, sq_type=None, question=None,
                             cycle=None):
        """
        Get survey questions.

        `Args:`
            statuses: list
                Filter to a list of statuses of survey questions. One or more of ``Active``,
                ``Archived``, and ``Inactive``.
            name: str
                Filter to survey questions with names begin with the input.
            type: str
                Filter to survey questions of a given type.
            question: str
                Filter to survey questions with script questions that contain the given input.
            cycle: str
                Filter to survey suestions with the given cycle. A year in the format "YYYY".
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {'statuses': statuses,
                  '$top': self.page_size,
                  'name': name,
                  'type': sq_type,
                  'question': question,
                  'cycle': cycle}

        tbl = Table(self.connection.get_request('surveyQuestions', params=params))
        logger.info(f'Found {tbl.num_rows} survey questions.')
        return tbl

    def get_survey_question(self, survey_question_id):
        """
        Get a survey question.

        `Args:`
            survey_question_id: int
                The survey question id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.connection.get_request(f'surveyQuestions/{survey_question_id}')
        logger.info(f'Found survey question {survey_question_id}.')
        return r

    def apply_survey_response(self, id, survey_question_id, survey_response_id,
                              id_type='vanid', result_code_id=None, contact_type_id=None,
                              input_type_id=None, date_canvassed=None):
        """
        Apply a single survey response to a person.

        `Args:`
            id: str
                A valid person id
            survey_question_id: int
                A valid survey question id
            survey_response_id: int
                A valid survey response id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
        """

        response = {"surveyQuestionId": survey_question_id,
                    "surveyResponseId": survey_response_id,
                    "type": "surveyResponse"}

        logger.info(f'Applying survey question {survey_question_id} to {id_type} {id}')
        self.apply_response(id, response, id_type, result_code_id=result_code_id,
                            contact_type_id=contact_type_id, input_type_id=input_type_id,
                            date_canvassed=date_canvassed)
